@@ -87,9 +87,9 @@ ucp_am_eager_short_proto_progress_common(uct_pending_req_t *self, int is_reply)
     return UCS_OK;
 }
 
-static ucs_status_t
-ucp_am_eager_short_proto_init_common(const ucp_proto_init_params_t *init_params,
-                                     ucp_operation_id_t op_id)
+static void
+ucp_am_eager_short_probe_common(const ucp_proto_init_params_t *init_params,
+                                ucp_operation_id_t op_id)
 {
     const ucp_proto_select_param_t *select_param = init_params->select_param;
     ucp_proto_single_init_params_t params        = {
@@ -111,6 +111,7 @@ ucp_am_eager_short_proto_init_common(const ucp_proto_init_params_t *init_params,
                                UCP_PROTO_COMMON_INIT_FLAG_CAP_SEG_SIZE |
                                UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
         .super.exclude_map   = 0,
+        .super.reg_mem_info  = ucp_mem_info_unknown,
         .lane_type           = UCP_LANE_TYPE_AM,
         .tl_cap_flags        = UCT_IFACE_FLAG_AM_SHORT
     };
@@ -118,16 +119,15 @@ ucp_am_eager_short_proto_init_common(const ucp_proto_init_params_t *init_params,
     if (!ucp_am_check_init_params(init_params, UCS_BIT(op_id),
                                   UCP_PROTO_SELECT_OP_FLAG_AM_RNDV) ||
         !ucp_proto_is_short_supported(select_param)) {
-        return UCS_ERR_UNSUPPORTED;
+        return;
     }
 
-    return ucp_proto_single_init(&params);
+    ucp_proto_single_probe(&params);
 }
 
-static ucs_status_t
-ucp_am_eager_short_proto_init(const ucp_proto_init_params_t *init_params)
+static void ucp_am_eager_short_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_short_proto_init_common(init_params, UCP_OP_ID_AM_SEND);
+    ucp_am_eager_short_probe_common(init_params, UCP_OP_ID_AM_SEND);
 }
 
 static ucs_status_t ucp_am_eager_short_proto_progress(uct_pending_req_t *self)
@@ -139,18 +139,17 @@ ucp_proto_t ucp_am_eager_short_proto = {
     .name     = "am/egr/short",
     .desc     = UCP_PROTO_SHORT_DESC,
     .flags    = UCP_PROTO_FLAG_AM_SHORT,
-    .init     = ucp_am_eager_short_proto_init,
+    .probe    = ucp_am_eager_short_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_short_proto_progress},
     .abort    = ucp_proto_am_request_bcopy_abort,
     .reset    = ucp_proto_request_bcopy_reset
 };
 
-static ucs_status_t
-ucp_am_eager_short_reply_proto_init(const ucp_proto_init_params_t *init_params)
+static void
+ucp_am_eager_short_reply_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_short_proto_init_common(init_params,
-                                                UCP_OP_ID_AM_SEND_REPLY);
+    ucp_am_eager_short_probe_common(init_params, UCP_OP_ID_AM_SEND_REPLY);
 }
 
 static ucs_status_t
@@ -163,7 +162,7 @@ ucp_proto_t ucp_am_eager_short_reply_proto = {
     .name     = "am/egr/short/reply",
     .desc     = UCP_PROTO_SHORT_DESC,
     .flags    = UCP_PROTO_FLAG_AM_SHORT,
-    .init     = ucp_am_eager_short_reply_proto_init,
+    .probe    = ucp_am_eager_short_reply_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_short_reply_proto_progress},
     .abort    = ucp_proto_am_request_bcopy_abort,
@@ -219,14 +218,14 @@ ucp_am_eager_single_bcopy_proto_progress(uct_pending_req_t *self)
     return ucp_proto_am_handle_user_header_send_status(req, status);
 }
 
-static ucs_status_t ucp_am_eager_single_bcopy_proto_init_common(
+static void ucp_am_eager_single_bcopy_probe_common(
         const ucp_proto_init_params_t *init_params, ucp_operation_id_t op_id)
 {
     ucp_context_t *context                = init_params->worker->context;
     ucp_proto_single_init_params_t params = {
         .super.super         = *init_params,
         .super.latency       = 0,
-        .super.overhead      = 5e-9,
+        .super.overhead      = context->config.ext.proto_overhead_single,
         .super.cfg_thresh    = context->config.ext.bcopy_thresh,
         .super.cfg_priority  = 20,
         .super.min_length    = 0,
@@ -242,41 +241,41 @@ static ucs_status_t ucp_am_eager_single_bcopy_proto_init_common(
                                UCP_PROTO_COMMON_INIT_FLAG_CAP_SEG_SIZE |
                                UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
         .super.exclude_map   = 0,
+        .super.reg_mem_info  = ucp_mem_info_unknown,
         .lane_type           = UCP_LANE_TYPE_AM,
         .tl_cap_flags        = UCT_IFACE_FLAG_AM_BCOPY
     };
 
     if (!ucp_am_check_init_params(init_params, UCS_BIT(op_id),
                                   UCP_PROTO_SELECT_OP_FLAG_AM_RNDV)) {
-        return UCS_ERR_UNSUPPORTED;
+        return;
     }
 
-    return ucp_proto_single_init(&params);
+    ucp_proto_single_probe(&params);
 }
 
-static ucs_status_t
-ucp_am_eager_single_bcopy_proto_init(const ucp_proto_init_params_t *init_params)
+static void
+ucp_am_eager_single_bcopy_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_single_bcopy_proto_init_common(init_params,
-                                                       UCP_OP_ID_AM_SEND);
+    ucp_am_eager_single_bcopy_probe_common(init_params, UCP_OP_ID_AM_SEND);
 }
 
 ucp_proto_t ucp_am_eager_single_bcopy_proto = {
     .name     = "am/egr/single/bcopy",
     .desc     = UCP_PROTO_COPY_IN_DESC,
     .flags    = 0,
-    .init     = ucp_am_eager_single_bcopy_proto_init,
+    .probe    = ucp_am_eager_single_bcopy_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_single_bcopy_proto_progress},
     .abort    = ucp_proto_am_request_bcopy_abort,
     .reset    = ucp_proto_request_bcopy_reset
 };
 
-static ucs_status_t ucp_am_eager_single_bcopy_reply_proto_init(
-        const ucp_proto_init_params_t *init_params)
+static void
+ucp_am_eager_single_bcopy_reply_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_single_bcopy_proto_init_common(init_params,
-                                                       UCP_OP_ID_AM_SEND_REPLY);
+    ucp_am_eager_single_bcopy_probe_common(init_params,
+                                           UCP_OP_ID_AM_SEND_REPLY);
 }
 
 static size_t ucp_am_eager_single_bcopy_reply_pack(void *dest, void *arg)
@@ -302,14 +301,14 @@ ucp_proto_t ucp_am_eager_single_bcopy_reply_proto = {
     .name     = "am/egr/single/bcopy/reply",
     .desc     = UCP_PROTO_COPY_IN_DESC,
     .flags    = 0,
-    .init     = ucp_am_eager_single_bcopy_reply_proto_init,
+    .probe    = ucp_am_eager_single_bcopy_reply_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_single_bcopy_reply_proto_progress},
     .abort    = ucp_proto_am_request_bcopy_abort,
     .reset    = ucp_proto_request_bcopy_reset
 };
 
-static ucs_status_t ucp_am_eager_single_zcopy_proto_init_common(
+static void ucp_am_eager_single_zcopy_probe_common(
         const ucp_proto_init_params_t *init_params, ucp_operation_id_t op_id)
 {
     ucp_context_t *context                = init_params->worker->context;
@@ -333,6 +332,8 @@ static ucs_status_t ucp_am_eager_single_zcopy_proto_init_common(
                                UCP_PROTO_COMMON_INIT_FLAG_CAP_SEG_SIZE |
                                UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
         .super.exclude_map   = 0,
+        .super.reg_mem_info  = ucp_proto_common_select_param_mem_info(
+                                                     init_params->select_param),
         .lane_type           = UCP_LANE_TYPE_AM,
         .tl_cap_flags        = UCT_IFACE_FLAG_AM_ZCOPY
     };
@@ -340,17 +341,16 @@ static ucs_status_t ucp_am_eager_single_zcopy_proto_init_common(
     if (!ucp_am_check_init_params(init_params, UCS_BIT(op_id),
                                   UCP_PROTO_SELECT_OP_FLAG_AM_RNDV) ||
         (init_params->select_param->dt_class != UCP_DATATYPE_CONTIG)) {
-        return UCS_ERR_UNSUPPORTED;
+        return;
     }
 
-    return ucp_proto_single_init(&params);
+    ucp_proto_single_probe(&params);
 }
 
-static ucs_status_t
-ucp_am_eager_single_zcopy_proto_init(const ucp_proto_init_params_t *init_params)
+static void
+ucp_am_eager_single_zcopy_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_single_zcopy_proto_init_common(init_params,
-                                                       UCP_OP_ID_AM_SEND);
+    ucp_am_eager_single_zcopy_probe_common(init_params, UCP_OP_ID_AM_SEND);
 }
 
 static ucs_status_t
@@ -402,18 +402,18 @@ ucp_proto_t ucp_am_eager_single_zcopy_proto = {
     .name     = "am/egr/single/zcopy",
     .desc     = UCP_PROTO_ZCOPY_DESC,
     .flags    = 0,
-    .init     = ucp_am_eager_single_zcopy_proto_init,
+    .probe    = ucp_am_eager_single_zcopy_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_single_zcopy_proto_progress},
     .abort    = ucp_proto_am_request_zcopy_abort,
     .reset    = ucp_am_proto_request_zcopy_reset
 };
 
-static ucs_status_t ucp_am_eager_single_zcopy_reply_proto_init(
-        const ucp_proto_init_params_t *init_params)
+static void
+ucp_am_eager_single_zcopy_reply_probe(const ucp_proto_init_params_t *init_params)
 {
-    return ucp_am_eager_single_zcopy_proto_init_common(init_params,
-                                                       UCP_OP_ID_AM_SEND_REPLY);
+    ucp_am_eager_single_zcopy_probe_common(init_params,
+                                           UCP_OP_ID_AM_SEND_REPLY);
 }
 
 static ucs_status_t
@@ -458,7 +458,7 @@ ucp_proto_t ucp_am_eager_single_zcopy_reply_proto = {
     .name     = "am/egr/single/zcopy/reply",
     .desc     = UCP_PROTO_ZCOPY_DESC,
     .flags    = 0,
-    .init     = ucp_am_eager_single_zcopy_reply_proto_init,
+    .probe    = ucp_am_eager_single_zcopy_reply_probe,
     .query    = ucp_proto_single_query,
     .progress = {ucp_am_eager_single_zcopy_reply_proto_progress},
     .abort    = ucp_proto_am_request_zcopy_abort,
