@@ -1,6 +1,7 @@
 /**
  * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2019. ALL RIGHTS RESERVED.
  * Copyright (C) ARM Ltd. 2016.  ALL RIGHTS RESERVED.
+ * Copyright (C) Intel Corporation, 2023.  ALL RIGHTS RESERVED.
  *
  * See file LICENSE for terms.
  */
@@ -14,6 +15,7 @@
 
 #include <ucs/config/parser.h>
 #include <ucs/algorithm/crc.h>
+#include <ucs/arch/atomic.h>
 #include <ucs/datastruct/mpool.inl>
 #include <ucs/datastruct/queue.h>
 #include <ucs/datastruct/string_set.h>
@@ -59,7 +61,9 @@
 
 #define UCP_AM_HANDLER_ENTRY(_id) [_id] = &ucp_am_handler_##_id,
 
-#define UCP_CPU_EST_BCOPY_BW_DEFAULT (5800 * UCS_MBYTE)
+#define UCP_CPU_EST_BCOPY_BW_DEFAULT         (7000 * UCS_MBYTE)
+#define UCP_CPU_EST_BCOPY_BW_DEFAULT_PROTOV1 (5800 * UCS_MBYTE)
+#define UCP_CPU_EST_BCOPY_BW_AMD_PROTOV1     (5008 * UCS_MBYTE)
 
 #define UCP_TL_AUX_SUFFIX    "aux"
 #define UCP_TL_AUX(_tl_name) _tl_name ":" UCP_TL_AUX_SUFFIX
@@ -111,6 +115,9 @@ static size_t ucp_rndv_frag_default_sizes[] = {
     [UCS_MEMORY_TYPE_ROCM]         = 4 * UCS_MBYTE,
     [UCS_MEMORY_TYPE_ROCM_MANAGED] = 4 * UCS_MBYTE,
     [UCS_MEMORY_TYPE_RDMA]         = 0,
+    [UCS_MEMORY_TYPE_ZE_HOST]      = 4 * UCS_MBYTE,
+    [UCS_MEMORY_TYPE_ZE_DEVICE]    = 4 * UCS_MBYTE,
+    [UCS_MEMORY_TYPE_ZE_MANAGED]   = 4 * UCS_MBYTE,
     [UCS_MEMORY_TYPE_LAST]         = 0
 };
 
@@ -121,6 +128,9 @@ static size_t ucp_rndv_frag_default_num_elems[] = {
     [UCS_MEMORY_TYPE_ROCM]         = 128,
     [UCS_MEMORY_TYPE_ROCM_MANAGED] = 128,
     [UCS_MEMORY_TYPE_RDMA]         = 0,
+    [UCS_MEMORY_TYPE_ZE_HOST]      = 128,
+    [UCS_MEMORY_TYPE_ZE_DEVICE]    = 128,
+    [UCS_MEMORY_TYPE_ZE_MANAGED]   = 128,
     [UCS_MEMORY_TYPE_LAST]         = 0
 };
 
@@ -130,15 +140,6 @@ const char *ucp_object_versions[] = {
     [UCP_OBJECT_VERSION_LAST] = NULL
 };
 
-const size_t ucp_context_est_bcopy_bw[UCS_CPU_VENDOR_LAST] = {
-    [UCS_CPU_VENDOR_UNKNOWN]     = UCP_CPU_EST_BCOPY_BW_DEFAULT,
-    [UCS_CPU_VENDOR_INTEL]       = UCP_CPU_EST_BCOPY_BW_DEFAULT,
-    [UCS_CPU_VENDOR_AMD]         = UCS_CPU_EST_BCOPY_BW_AMD,
-    [UCS_CPU_VENDOR_GENERIC_ARM] = UCP_CPU_EST_BCOPY_BW_DEFAULT,
-    [UCS_CPU_VENDOR_GENERIC_PPC] = UCP_CPU_EST_BCOPY_BW_DEFAULT,
-    [UCS_CPU_VENDOR_FUJITSU_ARM] = UCS_CPU_EST_BCOPY_BW_FUJITSU_ARM,
-    [UCS_CPU_VENDOR_ZHAOXIN]     = UCP_CPU_EST_BCOPY_BW_DEFAULT
-};
 
 static UCS_CONFIG_DEFINE_ARRAY(memunit_sizes, sizeof(size_t),
                                UCS_CONFIG_TYPE_MEMUNITS);
@@ -150,7 +151,7 @@ static ucs_config_field_t ucp_context_config_table[] = {
 
   {"MEMTYPE_REG_WHOLE_ALLOC_TYPES", "cuda",
    "Memory types which have whole allocations registered.\n"
-   "Allowed memory types: cuda, rocm, rocm-managed",
+   "Allowed memory types: cuda, rocm, rocm-managed, ze-host, ze-device, ze-managed",
    ucs_offsetof(ucp_context_config_t, reg_whole_alloc_bitmap),
    UCS_CONFIG_TYPE_BITMAP(ucs_memory_type_names)},
 
@@ -164,8 +165,14 @@ static ucs_config_field_t ucp_context_config_table[] = {
    ucs_offsetof(ucp_context_config_t, bcopy_thresh), UCS_CONFIG_TYPE_MEMUNITS},
 
   {"RNDV_THRESH", UCS_VALUE_AUTO_STR,
-   "Threshold for switching from eager to rendezvous protocol",
-   ucs_offsetof(ucp_context_config_t, rndv_thresh), UCS_CONFIG_TYPE_MEMUNITS},
+   "Threshold for switching from eager to rendezvous protocol", 0,
+    UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_MEMUNITS,
+        {"intra", "threshold for intra-node communication",
+         ucs_offsetof(ucp_context_config_t, rndv_intra_thresh)},
+        {"inter", "threshold for inter-node communication",
+         ucs_offsetof(ucp_context_config_t, rndv_inter_thresh)},
+        {NULL}
+  )},
 
   {"RNDV_SEND_NBR_THRESH", "256k",
    "Threshold for switching from eager to rendezvous protocol in ucp_tag_send_nbr().\n"
@@ -221,6 +228,10 @@ static ucs_config_field_t ucp_context_config_table[] = {
    "multiple rails. Must be greater than 0.",
    ucs_offsetof(ucp_context_config_t, min_rndv_chunk_size), UCS_CONFIG_TYPE_MEMUNITS},
 
+  {"RMA_ZCOPY_MAX_SEG_SIZE", "auto",
+   "Max size of a segment for rma/rndv zcopy.",
+   ucs_offsetof(ucp_context_config_t, rma_zcopy_max_seg_size), UCS_CONFIG_TYPE_MEMUNITS},
+
   {"RNDV_SCHEME", "auto",
    "Communication scheme in RNDV protocol.\n"
    " get_zcopy - use get_zcopy scheme in RNDV protocol.\n"
@@ -329,11 +340,15 @@ static ucs_config_field_t ucp_context_config_table[] = {
    "and the resulting performance.",
    ucs_offsetof(ucp_context_config_t, estimated_num_ppn), UCS_CONFIG_TYPE_ULUNITS},
 
-  {"RNDV_FRAG_MEM_TYPE", "host",
-   "Memory type of fragments used for RNDV pipeline protocol.\n"
-   "Allowed memory types is one of: host, cuda, rocm",
-   ucs_offsetof(ucp_context_config_t, rndv_frag_mem_type),
-   UCS_CONFIG_TYPE_ENUM(ucs_memory_type_names)},
+  {"RNDV_FRAG_MEM_TYPE", NULL, "",
+   ucs_offsetof(ucp_context_config_t, rndv_frag_mem_types),
+   UCS_CONFIG_TYPE_BITMAP(ucs_memory_type_names)},
+
+  {"RNDV_FRAG_MEM_TYPES", "host,cuda",
+   "Memory types of fragments used for RNDV pipeline protocol.\n"
+   "Allowed memory types are: host, cuda, rocm, ze-host, ze-device",
+   ucs_offsetof(ucp_context_config_t, rndv_frag_mem_types),
+   UCS_CONFIG_TYPE_BITMAP(ucs_memory_type_names)},
 
   {"RNDV_PIPELINE_SEND_THRESH", "inf",
    "RNDV size threshold to enable sender side pipeline for mem type",
@@ -343,6 +358,11 @@ static ucs_config_field_t ucp_context_config_table[] = {
    "Use two stage pipeline rendezvous protocol for intra-node GPU to GPU transfers",
    ucs_offsetof(ucp_context_config_t, rndv_shm_ppln_enable), UCS_CONFIG_TYPE_BOOL},
 
+  {"RNDV_PIPELINE_ERROR_HANDLING", "n",
+   "Allow using error handling protocol in the rendezvous pipeline protocol\n"
+   "even if invalidation workflow isn't supported",
+   ucs_offsetof(ucp_context_config_t, rndv_errh_ppln_enable), UCS_CONFIG_TYPE_BOOL},
+
   {"FLUSH_WORKER_EPS", "y",
    "Enable flushing the worker by flushing its endpoints. Allows completing\n"
    "the flush operation in a bounded time even if there are new requests on\n"
@@ -394,6 +414,18 @@ static ucs_config_field_t ucp_context_config_table[] = {
    "(inf - check all endpoints on every round, must be greater than 0)",
    ucs_offsetof(ucp_context_config_t, keepalive_num_eps), UCS_CONFIG_TYPE_UINT},
 
+  {"DYNAMIC_TL_SWITCH_INTERVAL", "inf",
+   "Time interval between dynamic transport switching rounds. Must be\n"
+   "non-zero value. use 'inf' to disable this feature.",
+   ucs_offsetof(ucp_context_config_t, dynamic_tl_switch_interval),
+   UCS_CONFIG_TYPE_TIME_UNITS},
+
+  {"DYNAMIC_TL_PROGRESS_FACTOR", "10",
+   "Number of usage tracker rounds performed for each progress operation. Must be\n"
+   "non-zero value.",
+   ucs_offsetof(ucp_context_config_t, dynamic_tl_progress_factor),
+   UCS_CONFIG_TYPE_TIME_UNITS},
+
   {"RESOLVE_REMOTE_EP_ID", "n",
    "Defines whether resolving remote endpoint ID is required or not when\n"
    "creating a local endpoint. 'auto' means resolving remote endpoint ID only\n"
@@ -452,8 +484,9 @@ static ucs_config_field_t ucp_context_config_table[] = {
    ucs_offsetof(ucp_context_config_t, proto_info_dir), UCS_CONFIG_TYPE_STRING},
 
   {"REG_NONBLOCK_MEM_TYPES", "",
-   "Perform only non-blocking memory registration for these memory types:\n"
-   "page registration may be deferred until it is accessed by the CPU or a transport.",
+   "Perform only non-blocking memory registration for these memory types.\n"
+   "Non-blocking registration means that the page registration may be\n"
+   "deferred until it is accessed by the CPU or a transport.",
    ucs_offsetof(ucp_context_config_t, reg_nb_mem_types),
    UCS_CONFIG_TYPE_BITMAP(ucs_memory_type_names)},
 
@@ -464,6 +497,41 @@ static ucs_config_field_t ucp_context_config_table[] = {
    " 'n' : Select RMA/AMO lanes according to performance charasteristics",
    ucs_offsetof(ucp_context_config_t, prefer_offload), UCS_CONFIG_TYPE_BOOL},
 
+  {"PROTO_OVERHEAD", "single:5ns,multi:10ns,rndv_offload:40ns,rndv_rtr:40ns,"
+                     "rndv_rts:275ns,sw:40ns,rkey_ptr:0",
+   "Protocol overhead", 0,
+    UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME,
+        {"single", "overhead of single-lane protocol",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_single)},
+        {"multi", "overhead of managing multiple lanes",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_multi)},
+        {"rndv_offload", "overhead of rendezvous offload protocol",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_rndv_offload)},
+        {"rndv_rtr", "overhead of rendezvous RTR protocol",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_rndv_rtr)},
+        {"rndv_rts", "overhead of rendezvous RTS protocol",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_rndv_rts)},
+        {"sw", "overhead of software emulation protocol",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_sw)},
+        {"rkey_ptr", "overhead of the protocol copying from mapped remote "
+                     "memory",
+         ucs_offsetof(ucp_context_config_t, proto_overhead_rkey_ptr)},
+        {NULL}
+  )},
+
+  {"GVA_ENABLE", "off",
+   "Enable Global VA infrastructure. Setting to 'auto' will try to enable, "
+   "but if error handling enabled will disable",
+   ucs_offsetof(ucp_context_config_t, gva_enable), UCS_CONFIG_TYPE_ON_OFF_AUTO},
+
+  {"GVA_MLOCK", "y",
+   "Lock memory with mlock() when using global VA MR",
+   ucs_offsetof(ucp_context_config_t, gva_mlock), UCS_CONFIG_TYPE_BOOL},
+
+  {"GVA_PREFETCH", "y",
+   "Prefetch memory when using global VA MR",
+   ucs_offsetof(ucp_context_config_t, gva_prefetch), UCS_CONFIG_TYPE_BOOL},
+
   {NULL}
 };
 
@@ -505,6 +573,7 @@ static ucs_config_field_t ucp_config_table[] = {
    " - tcp     : sockets over TCP/IP.\n"
    " - cuda    : CUDA (NVIDIA GPU) memory support.\n"
    " - rocm    : ROCm (AMD GPU) memory support.\n"
+   " - ze      : ZE (Intel GPU) memory support.\n"
    " Using a \\ prefix before a transport name treats it as an explicit transport name\n"
    " and disables aliasing.",
    ucs_offsetof(ucp_config_t, tls), UCS_CONFIG_TYPE_ALLOW_LIST},
@@ -589,6 +658,7 @@ static ucp_tl_alias_t ucp_tl_aliases[] = {
   { "ugni",  { "ugni_smsg", UCP_TL_AUX("ugni_udt"), "ugni_rdma", NULL } },
   { "cuda",  { "cuda_copy", "cuda_ipc", "gdr_copy", NULL } },
   { "rocm",  { "rocm_copy", "rocm_ipc", "rocm_gdr", NULL } },
+  { "ze",    { "ze_copy", "ze_ipc", "ze_gdr", NULL } },
   { NULL }
 };
 
@@ -606,7 +676,7 @@ const char *ucp_feature_str[] = {
 
 
 const ucp_tl_bitmap_t ucp_tl_bitmap_max = {{UINT64_MAX, UINT64_MAX}};
-const ucp_tl_bitmap_t ucp_tl_bitmap_min = UCS_BITMAP_ZERO;
+const ucp_tl_bitmap_t ucp_tl_bitmap_min = {{0}};
 
 
 ucs_status_t ucp_config_read(const char *env_prefix, const char *filename,
@@ -720,7 +790,8 @@ void ucp_config_release(ucp_config_t *config)
 ucs_status_t ucp_config_modify_internal(ucp_config_t *config, const char *name,
                                         const char *value)
 {
-    return ucs_config_parser_set_value(config, ucp_config_table, name, value);
+    return ucs_config_parser_set_value(config, ucp_config_table, NULL, name,
+                                       value);
 }
 
 ucs_status_t ucp_config_modify(ucp_config_t *config, const char *name,
@@ -1238,7 +1309,7 @@ const char *ucp_tl_bitmap_str(ucp_context_h context,
     p    = str;
     endp = str + max_str_len;
 
-    UCS_BITMAP_FOR_EACH_BIT(*tl_bitmap, i) {
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(i, tl_bitmap) {
         ucs_snprintf_zero(p, endp - p, "%s ",
                           context->tl_rscs[i].tl_rsc.tl_name);
         p += strlen(p);
@@ -1247,13 +1318,15 @@ const char *ucp_tl_bitmap_str(ucp_context_h context,
     return str;
 }
 
-
 static void ucp_free_resources(ucp_context_t *context)
 {
     ucp_rsc_index_t i;
 
     ucs_free(context->tl_rscs);
     for (i = 0; i < context->num_mds; ++i) {
+        if (context->tl_mds[i].gva_mr != NULL) {
+            uct_md_mem_dereg(context->tl_mds[i].md, context->tl_mds[i].gva_mr);
+        }
         uct_md_close(context->tl_mds[i].md);
     }
     ucs_free(context->tl_mds);
@@ -1398,13 +1471,13 @@ static void ucp_fill_sockaddr_cms_prio_list(ucp_context_h context,
          * sockaddr CMs bitmap. Save the cmpt_idx for the client/server usage
          * later */
         cm_cmpts_bitmap_safe = cm_cmpts_bitmap;
-        UCS_BITMAP_FOR_EACH_BIT(cm_cmpts_bitmap_safe, cmpt_idx) {
+        UCS_STATIC_BITMAP_FOR_EACH_BIT(cmpt_idx, &cm_cmpts_bitmap_safe) {
             if (!strcmp(sockaddr_cm_names[cm_idx], "*") ||
                 !strncmp(sockaddr_cm_names[cm_idx],
                          context->tl_cmpts[cmpt_idx].attr.name,
                          UCT_COMPONENT_NAME_MAX)) {
                 context->config.cm_cmpt_idxs[context->config.num_cm_cmpts++] = cmpt_idx;
-                UCS_BITMAP_UNSET(cm_cmpts_bitmap, cmpt_idx);
+                UCS_STATIC_BITMAP_RESET(&cm_cmpts_bitmap, cmpt_idx);
             }
         }
     }
@@ -1524,12 +1597,20 @@ ucp_add_component_resources(ucp_context_h context, ucp_rsc_index_t cmpt_index,
             }
 
             ucs_memory_type_for_each(mem_type) {
-                if ((context->config.ext.reg_nb_mem_types & UCS_BIT(mem_type)) &&
-                    !(md_attr->reg_nonblock_mem_types & UCS_BIT(mem_type))) {
-                    continue;
-                }
-
                 if (md_attr->flags & UCT_MD_FLAG_REG) {
+                    if ((context->config.ext.reg_nb_mem_types & UCS_BIT(mem_type)) &&
+                        !(md_attr->reg_nonblock_mem_types & UCS_BIT(mem_type))) {
+                        if (md_attr->reg_mem_types & UCS_BIT(mem_type)) {
+                            /* Keep map of MDs supporting blocking registration
+                             * if non-blocking registration is requested for the
+                             * given memory type. In some cases blocking
+                             * registration maybe required anyway (e.g. internal
+                             * staging buffers for rndv pipeline protocols). */
+                            context->reg_block_md_map[mem_type] |= UCS_BIT(md_index);
+                        }
+                        continue;
+                    }
+
                     if (md_attr->reg_mem_types & UCS_BIT(mem_type)) {
                         context->reg_md_map[mem_type] |= UCS_BIT(md_index);
                     }
@@ -1537,6 +1618,11 @@ ucp_add_component_resources(ucp_context_h context, ucp_rsc_index_t cmpt_index,
                     if (md_attr->cache_mem_types & UCS_BIT(mem_type)) {
                         context->cache_md_map[mem_type] |= UCS_BIT(md_index);
                     }
+
+                    if ((context->config.ext.gva_enable != UCS_CONFIG_OFF) &&
+                        (md_attr->gva_mem_types & UCS_BIT(mem_type))) {
+                        context->gva_md_map[mem_type] |= UCS_BIT(md_index);
+                    }
                 }
             }
 
@@ -1654,14 +1740,16 @@ static ucs_status_t ucp_fill_resources(ucp_context_h context,
     context->export_md_map            = 0;
 
     ucs_memory_type_for_each(mem_type) {
-        context->reg_md_map[mem_type]     = 0;
-        context->cache_md_map[mem_type]   = 0;
-        context->dmabuf_mds[mem_type]     = UCP_NULL_RESOURCE;
-        context->alloc_md_index[mem_type] = UCP_NULL_RESOURCE;
+        context->reg_md_map[mem_type]           = 0;
+        context->reg_block_md_map[mem_type]     = 0;
+        context->cache_md_map[mem_type]         = 0;
+        context->gva_md_map[mem_type]           = 0;
+        context->dmabuf_mds[mem_type]           = UCP_NULL_RESOURCE;
+        context->alloc_md[mem_type].md_index    = UCP_NULL_RESOURCE;
+        context->alloc_md[mem_type].sys_dev     = UCS_SYS_DEVICE_ID_UNKNOWN;
+        context->alloc_md[mem_type].initialized = 0;
     }
 
-    context->alloc_md_index_initialized = 0;
-
     ucs_string_set_init(&avail_tls);
     UCS_STATIC_ASSERT(UCT_DEVICE_TYPE_NET == 0);
     for (dev_type = UCT_DEVICE_TYPE_NET; dev_type < UCT_DEVICE_TYPE_LAST; ++dev_type) {
@@ -1699,7 +1787,7 @@ static ucs_status_t ucp_fill_resources(ucp_context_h context,
         goto out_release_components;
     }
 
-    UCS_BITMAP_CLEAR(&context->config.cm_cmpts_bitmap);
+    UCS_STATIC_BITMAP_RESET_ALL(&context->config.cm_cmpts_bitmap);
 
     max_mds = 0;
     for (i = 0; i < context->num_cmpts; ++i) {
@@ -1715,14 +1803,14 @@ static ucs_status_t ucp_fill_resources(ucp_context_h context,
         }
 
         if (context->tl_cmpts[i].attr.flags & UCT_COMPONENT_FLAG_CM) {
-            UCS_BITMAP_SET(context->config.cm_cmpts_bitmap, i);
+            UCS_STATIC_BITMAP_SET(&context->config.cm_cmpts_bitmap, i);
         }
 
         max_mds += context->tl_cmpts[i].attr.md_resource_count;
     }
 
     /* Allocate actual array of MDs */
-    context->tl_mds = ucs_malloc(max_mds * sizeof(*context->tl_mds),
+    context->tl_mds = ucs_calloc(max_mds, sizeof(*context->tl_mds),
                                  "ucp_tl_mds");
     if (context->tl_mds == NULL) {
         status = UCS_ERR_NO_MEMORY;
@@ -1745,8 +1833,8 @@ static ucs_status_t ucp_fill_resources(ucp_context_h context,
      * Then the worker will open all available transport resources and will
      * select only the best ones for each particular device.
      */
-    UCS_BITMAP_MASK(&context->tl_bitmap,
-                    config->ctx.unified_mode ? 0 : context->num_tls);
+    UCS_STATIC_BITMAP_MASK(&context->tl_bitmap,
+                           config->ctx.unified_mode ? 0 : context->num_tls);
 
     /* Warn about devices and transports which were specified explicitly in the
      * configuration, but are not available
@@ -1761,8 +1849,11 @@ static ucs_status_t ucp_fill_resources(ucp_context_h context,
         }
 
         ucp_get_aliases_set(&avail_tls);
-        ucp_report_unavailable(&config->tls.array, tl_cfg_mask, "", "transport",
-                               &avail_tls);
+
+        if (config->tls.mode == UCS_CONFIG_ALLOW_LIST_ALLOW) {
+            ucp_report_unavailable(&config->tls.array, tl_cfg_mask, "", "transport",
+                                   &avail_tls);
+        }
     }
 
     /* Validate context resources */
@@ -1792,6 +1883,8 @@ err_free_resources:
 static void ucp_apply_params(ucp_context_h context, const ucp_params_t *params,
                              ucp_mt_type_t mt_type)
 {
+    static uint64_t context_counter = 0;
+
     context->config.features = UCP_PARAM_FIELD_VALUE(params, features, FEATURES,
                                                      0);
     if (!context->config.features) {
@@ -1831,7 +1924,8 @@ static void ucp_apply_params(ucp_context_h context, const ucp_params_t *params,
         ucs_snprintf_zero(context->name, UCP_ENTITY_NAME_MAX, "%s",
                           params->name);
     } else {
-        ucs_snprintf_zero(context->name, UCP_ENTITY_NAME_MAX, "%p", context);
+        ucs_snprintf_zero(context->name, UCP_ENTITY_NAME_MAX, "ucp_context_%lu",
+                          ucs_atomic_fadd64(&context_counter, 1));
     }
 }
 
@@ -1870,9 +1964,27 @@ ucp_fill_rndv_frag_config(const ucp_context_config_names_t *config,
     return UCS_OK;
 }
 
-static double ucp_context_get_memcpy_bw()
+static double ucp_context_get_protov1_memcpy_bw()
 {
-    return ucp_context_est_bcopy_bw[ucs_arch_get_cpu_vendor()];
+    return (ucs_arch_get_cpu_vendor() == UCS_CPU_VENDOR_AMD) ?
+                   UCP_CPU_EST_BCOPY_BW_AMD_PROTOV1 :
+                   UCP_CPU_EST_BCOPY_BW_DEFAULT_PROTOV1;
+}
+
+static int
+ucp_dynamic_tl_switch_config_valid(const ucp_context_config_t *config)
+{
+    if (config->dynamic_tl_switch_interval == 0) {
+        ucs_error("UCX_DYNAMIC_TL_SWITCH_INTERVAL must be > 0");
+        return 0;
+    }
+
+    if (config->dynamic_tl_progress_factor == 0) {
+        ucs_error("UCX_DYNAMIC_TL_PROGRESS_FACTOR must be > 0");
+        return 0;
+    }
+
+    return 1;
 }
 
 static ucs_status_t ucp_fill_config(ucp_context_h context,
@@ -1913,9 +2025,9 @@ static ucs_status_t ucp_fill_config(ucp_context_h context,
     if (UCS_CONFIG_DBL_IS_AUTO(context->config.ext.bcopy_bw)) {
         /* bcopy_bw wasn't set via the env variable. Calculate the value */
         if (context->config.ext.proto_enable) {
-            context->config.ext.bcopy_bw = ucs_cpu_get_memcpy_bw();
+            context->config.ext.bcopy_bw = UCP_CPU_EST_BCOPY_BW_DEFAULT;
         } else {
-            context->config.ext.bcopy_bw = ucp_context_get_memcpy_bw();
+            context->config.ext.bcopy_bw = ucp_context_get_protov1_memcpy_bw();
         }
     }
     ucs_debug("estimated bcopy bandwidth is %f", context->config.ext.bcopy_bw);
@@ -2046,6 +2158,11 @@ static ucs_status_t ucp_fill_config(ucp_context_h context,
         goto err_free_alloc_methods;
     }
 
+    if (!ucp_dynamic_tl_switch_config_valid(&context->config.ext)) {
+        status = UCS_ERR_INVALID_PARAM;
+        goto err_free_alloc_methods;
+    }
+
     ucs_list_for_each(key_val, &config->cached_key_list, list) {
         status = ucp_config_cached_key_add(&context->cached_key_list,
                                            key_val->key, key_val->value);
@@ -2071,6 +2188,9 @@ static ucs_status_t ucp_fill_config(ucp_context_h context,
              ((context->config.ext.max_rma_lanes > 1) ||
               context->config.ext.proto_enable));
 
+    context->config.progress_wrapper_enabled =
+            ucs_log_is_enabled(UCS_LOG_LEVEL_TRACE_REQ) ||
+            ucp_context_usage_tracker_enabled(context);
     return UCS_OK;
 
 err_free_key_list:
@@ -2414,13 +2534,13 @@ ucp_context_dev_tl_bitmap(ucp_context_h context, const char *dev_name,
 {
     ucp_rsc_index_t tl_idx;
 
-    UCS_BITMAP_CLEAR(tl_bitmap);
-    UCS_BITMAP_FOR_EACH_BIT(context->tl_bitmap, tl_idx) {
+    UCS_STATIC_BITMAP_RESET_ALL(tl_bitmap);
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(tl_idx, &context->tl_bitmap) {
         if (strcmp(context->tl_rscs[tl_idx].tl_rsc.dev_name, dev_name)) {
             continue;
         }
 
-        UCS_BITMAP_SET(*tl_bitmap, tl_idx);
+        UCS_STATIC_BITMAP_SET(tl_bitmap, tl_idx);
     }
 }
 
@@ -2430,10 +2550,10 @@ ucp_context_dev_idx_tl_bitmap(ucp_context_h context, ucp_rsc_index_t dev_idx,
 {
     ucp_rsc_index_t tl_idx;
 
-    UCS_BITMAP_CLEAR(tl_bitmap);
-    UCS_BITMAP_FOR_EACH_BIT(context->tl_bitmap, tl_idx) {
+    UCS_STATIC_BITMAP_RESET_ALL(tl_bitmap);
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(tl_idx, &context->tl_bitmap) {
         if (context->tl_rscs[tl_idx].dev_index == dev_idx) {
-            UCS_BITMAP_SET(*tl_bitmap, tl_idx);
+            UCS_STATIC_BITMAP_SET(tl_bitmap, tl_idx);
         }
     }
 }
@@ -2441,9 +2561,10 @@ ucp_context_dev_idx_tl_bitmap(ucp_context_h context, ucp_rsc_index_t dev_idx,
 void ucp_tl_bitmap_validate(const ucp_tl_bitmap_t *tl_bitmap,
                             const ucp_tl_bitmap_t *tl_bitmap_super)
 {
-    ucs_assert_always(UCS_BITMAP_IS_ZERO(UCP_TL_BITMAP_AND_NOT(*tl_bitmap,
-                                                               *tl_bitmap_super),
-                                         UCP_MAX_RESOURCES));
+    ucp_tl_bitmap_t b = UCS_STATIC_BITMAP_AND(*tl_bitmap,
+                                              UCS_STATIC_BITMAP_NOT(
+                                                      *tl_bitmap_super));
+    ucs_assert_always(UCS_STATIC_BITMAP_IS_ZERO(b));
 }
 
 const char* ucp_context_cm_name(ucp_context_h context, ucp_rsc_index_t cm_idx)
