@@ -23,6 +23,7 @@
 #include <ucs/datastruct/conn_match.h>
 #include <ucs/memory/memtype_cache.h>
 #include <ucs/memory/memory_type.h>
+#include <ucs/memory/rcache.h>
 #include <ucs/type/spinlock.h>
 #include <ucs/sys/string.h>
 #include <ucs/type/param.h>
@@ -78,12 +79,14 @@ typedef struct ucp_context_config {
     size_t                                 rndv_frag_size[UCS_MEMORY_TYPE_LAST];
     /** Number of RNDV pipeline fragments per allocation */
     size_t                                 rndv_num_frags[UCS_MEMORY_TYPE_LAST];
-    /** Memory type of fragments used for RNDV pipeline protocol */
-    ucs_memory_type_t                      rndv_frag_mem_type;
+    /** Memory types of fragments used for RNDV pipeline protocol */
+    uint64_t                               rndv_frag_mem_types;
     /** RNDV pipeline send threshold */
     size_t                                 rndv_pipeline_send_thresh;
     /** Enabling 2-stage pipeline rndv protocol */
     int                                    rndv_shm_ppln_enable;
+    /** Enable error handling for rndv pipeline protocol */
+    int                                    rndv_errh_ppln_enable;
     /** Threshold for using tag matching offload capabilities. Smaller buffers
      *  will not be posted to the transport. */
     size_t                                 tm_thresh;
@@ -137,13 +140,17 @@ typedef struct ucp_context_config {
     /** Maximal number of endpoints to check on every keepalive round
      * (0 - disabled, inf - check all endpoints on every round) */
     unsigned                               keepalive_num_eps;
+    /** Time period between dynamic transport switching rounds */
+    ucs_time_t                             dynamic_tl_switch_interval;
+    /** Number of usage tracker rounds performed for each progress operation */
+    unsigned                               dynamic_tl_progress_factor;
     /** Defines whether resolving remote endpoint ID is required or not when
      *  creating a local endpoint */
     ucs_on_off_auto_value_t                resolve_remote_ep_id;
     /** Enable indirect IDs to object pointers in wire protocols */
     ucs_on_off_auto_value_t                proto_indirect_id;
     /** Bitmap of memory types whose allocations are registered fully */
-    unsigned                               reg_whole_alloc_bitmap;
+    uint64_t                               reg_whole_alloc_bitmap;
     /** Always use flush operation in rendezvous put */
     int                                    rndv_put_force_flush;
     /** Maximum size of mem type direct rndv*/
@@ -169,6 +176,12 @@ typedef struct ucp_context_config {
     int                                    prefer_offload;
     /** RMA zcopy segment size */
     size_t                                 rma_zcopy_max_seg_size;
+    /** Enable global VA MR */
+    ucs_on_off_auto_value_t                gva_enable;
+    /** Lock memory when using global VA MR */
+    int                                    gva_mlock;
+    /** Prefetch memory when using global VA MR */
+    int                                    gva_prefetch;
     /** Protocol overhead */
     double                                 proto_overhead_single;
     double                                 proto_overhead_multi;
@@ -176,6 +189,9 @@ typedef struct ucp_context_config {
     double                                 proto_overhead_rndv_rts;
     double                                 proto_overhead_rndv_rtr;
     double                                 proto_overhead_sw;
+    double                                 proto_overhead_rkey_ptr;
+    /** Registration cache lookup overhead estimation */
+    double                                 rcache_overhead;
 } ucp_context_config_t;
 
 
@@ -275,6 +291,11 @@ typedef struct ucp_tl_md {
      * Flags mask parameter for @ref uct_md_mkey_pack_v2
      */
     unsigned               pack_flags_mask;
+
+    /**
+     * Global VA memory handle
+     */
+    uct_mem_h              gva_mr;
 } ucp_tl_md_t;
 
 
@@ -282,7 +303,8 @@ typedef struct ucp_context_alloc_md_index {
     int            initialized;
     /* Index of memory domain that is used to allocate memory of the given type
      * using ucp_memh_alloc(). */
-    ucp_md_index_t md_index;
+    ucp_md_index_t   md_index;
+    ucs_sys_device_t sys_dev;
 } ucp_context_alloc_md_index_t;
 
 
@@ -312,6 +334,9 @@ typedef struct ucp_context {
     /* Map of MDs that require caching registrations for given memory type. */
     ucp_md_map_t                  cache_md_map[UCS_MEMORY_TYPE_LAST];
 
+    /* Map of MDs that support global VA MRs for given memory type. */
+    ucp_md_map_t                  gva_md_map[UCS_MEMORY_TYPE_LAST];
+
     /* Map of MDs that provide registration of a memory buffer for a given
        memory type to be exported to other processes. */
     ucp_md_map_t                  export_md_map;
@@ -388,6 +413,9 @@ typedef struct ucp_context {
         /* worker_fence implementation method */
         unsigned                  worker_strong_fence;
 
+        /* Progress wrapper enabled */
+        int                       progress_wrapper_enabled;
+
         struct {
            unsigned               count;
            size_t                 *sizes;
@@ -684,6 +712,12 @@ ucp_memory_detect(ucp_context_h context, const void *address, size_t length,
     mem_info->sys_dev = mem_info_internal.sys_dev;
 }
 
+static UCS_F_ALWAYS_INLINE int
+ucp_context_usage_tracker_enabled(ucp_context_h context)
+{
+    return context->config.ext.dynamic_tl_switch_interval != UCS_TIME_INFINITY;
+}
+
 void
 ucp_context_dev_tl_bitmap(ucp_context_h context, const char *dev_name,
                           ucp_tl_bitmap_t *tl_bitmap);
