@@ -22,10 +22,6 @@
 #define UCP_PROTO_MAX_PERF_RANGES 24
 
 
-/* Maximal size of protocol private data */
-#define UCP_PROTO_PRIV_MAX          1024
-
-
 /* Maximal number of protocols in total */
 #define UCP_PROTO_MAX_COUNT         64
 
@@ -88,90 +84,6 @@ enum {
 };
 
 
-/*
- * Some protocols can be pipelined, so the time they consume when multiple
- * such operations are issued is less than their cumulative time. Therefore we
- * define two metrics: "single" operation time and "multi" operation time.
- *
- * -------time---------->
- *
- *        +-------------------------+
- * op1:   |   "single" time         |
- *        +---------------+---------+---------------+
- *                op2:    | overlap | "multi" time  |
- *                        +---------+-----+---------+---------------+
- *                                op3:    | overlap | "multi" time  |
- *                                        +---------+---------------+
- */
-typedef enum {
-    UCP_PROTO_PERF_TYPE_FIRST,
-
-    /* Time to complete this operation assuming it's the only one. */
-    UCP_PROTO_PERF_TYPE_SINGLE = UCP_PROTO_PERF_TYPE_FIRST,
-
-    /* Time to complete this operation after all previous ones complete. */
-    UCP_PROTO_PERF_TYPE_MULTI,
-
-    /* CPU time the operation consumes (it would be less than or equal to the
-     * SINGLE and MULTI times).
-     */
-    UCP_PROTO_PERF_TYPE_CPU,
-
-    UCP_PROTO_PERF_TYPE_LAST
-} ucp_proto_perf_type_t;
-
-
-/*
- * Iterate over performance types.
- *
- * @param _perf_type  Performance type iterator variable.
- */
-#define UCP_PROTO_PERF_TYPE_FOREACH(_perf_type) \
-    for (_perf_type = UCP_PROTO_PERF_TYPE_FIRST; \
-         _perf_type < UCP_PROTO_PERF_TYPE_LAST; ++(_perf_type))
-
-
-/*
- * Performance estimation for a range of message sizes.
- */
-typedef struct {
-    /* Maximal payload size for this range */
-    size_t                max_length;
-
-    /* Estimated time in seconds, as a function of message size in bytes, to
-     * complete the operation. See @ref ucp_proto_perf_type_t for details
-     */
-    ucs_linear_func_t     perf[UCP_PROTO_PERF_TYPE_LAST];
-
-    /* Performance data tree */
-    ucp_proto_perf_node_t *node;
-} ucp_proto_perf_range_t;
-
-
-/**
- * UCP protocol capabilities (per operation parameters)
- */
-typedef struct {
-    /* Configured protocol threshold
-       NOTE will be removed when all protocols implement probe() instead of init() */
-    size_t                  cfg_thresh;
-
-    /* Priority of configuration
-       NOTE will be removed when all protocols implement probe() instead of init() */
-    unsigned                cfg_priority;
-
-    /* Minimal message size */
-    size_t                  min_length;
-
-    /* Number of entries in 'ranges' */
-    unsigned                num_ranges;
-
-    /* Performance estimation function for different message sizes */
-    ucp_proto_perf_range_t  ranges[UCP_PROTO_MAX_PERF_RANGES];
-
-} ucp_proto_caps_t;
-
-
 /**
  * Parameters for protocol initialization function
  */
@@ -186,40 +98,10 @@ typedef struct {
     const ucp_rkey_config_key_t    *rkey_config_key; /* Remote key configuration,
                                                         may be NULL */
     ucp_proto_id_t                 proto_id;         /* Initial protocol ID */
-    const char                     *proto_name;      /* Name of the initialized
-                                                        protocol, for debugging */
-    ucp_proto_probe_ctx_t          *ctx;             /* Selection context for
-                                                        adding a protocol instance */
-
-    /* Output parameters */
-    void                           *priv;       /* Pointer to priv buffer */
-    size_t                         *priv_size;  /* Occupied size in priv buffer */
-    ucp_proto_caps_t               *caps;       /* Protocol capabilities */
+    ucp_proto_probe_ctx_t          *ctx;             /* Context for adding caps */
 } ucp_proto_init_params_t;
 
 
-/**
- * Initialize protocol-specific configuration and estimate protocol performance
- * as function of message size.
- *
- * @param [in]  params   Protocol initialization parameters.
- */
-typedef void (*ucp_proto_probe_func_t)(const ucp_proto_init_params_t *params);
-
-
-/**
- * Initialize protocol-specific configuration and estimate protocol performance
- * as function of message size.
- *
- * @param [in]  params   Protocol initialization parameters.
- *
- * @return UCS_OK              - if successful.
- *         UCS_ERR_UNSUPPORTED - if the protocol is not supported on the key.
- */
-typedef ucs_status_t
-(*ucp_proto_init_func_t)(const ucp_proto_init_params_t *params);
-
-
 typedef struct {
     /* Protocol definition */
     const ucp_proto_t              *proto;
@@ -262,6 +144,15 @@ typedef struct {
 } ucp_proto_query_attr_t;
 
 
+/**
+ * Initialize protocol-specific configuration and estimate protocol performance
+ * as function of message size.
+ *
+ * @param [in]  params  Protocol initialization parameters.
+ */
+typedef void (*ucp_proto_probe_func_t)(const ucp_proto_init_params_t *params);
+
+
 /**
  * Query protocol-specific information.
  *
@@ -310,12 +201,6 @@ struct ucp_proto {
     /* Probe and add protocol instances */
     ucp_proto_probe_func_t   probe;
 
-    /*
-     * Initialization function
-     * TODO remove after all protocols are using probe
-     */
-    ucp_proto_init_func_t    init;
-
     /* Query protocol information */
     ucp_proto_query_func_t   query;
 
@@ -376,16 +261,4 @@ unsigned ucp_protocols_count(void);
 void ucp_proto_default_query(const ucp_proto_query_params_t *params,
                              ucp_proto_query_attr_t *attr);
 
-
-void ucp_proto_perf_set(ucs_linear_func_t perf[UCP_PROTO_PERF_TYPE_LAST],
-                        ucs_linear_func_t func);
-
-
-void ucp_proto_perf_copy(ucs_linear_func_t dest[UCP_PROTO_PERF_TYPE_LAST],
-                         const ucs_linear_func_t src[UCP_PROTO_PERF_TYPE_LAST]);
-
-
-void ucp_proto_perf_add(ucs_linear_func_t perf[UCP_PROTO_PERF_TYPE_LAST],
-                        ucs_linear_func_t func);
-
 #endif
