@@ -6,10 +6,15 @@
 #ifndef UCP_PROTO_DEBUG_H_
 #define UCP_PROTO_DEBUG_H_
 
+#include <ucs/datastruct/dynamic_bitmap.h>
+
 #include "proto_common.h"
 #include "proto_select.h"
 
 
+/* String to place new line inside performance graph */
+#define UCP_PROTO_PERF_NODE_NEW_LINE "<br/>"
+
 /* Format string to display a protocol time */
 #define UCP_PROTO_TIME_FMT(_time_var) " " #_time_var ": %.2f ns"
 #define UCP_PROTO_TIME_ARG(_time_val) ((_time_val) * 1e9)
@@ -25,24 +30,13 @@
     ((_perf_func)->m != 0.0) ? (1.0 / ((_perf_func)->m * UCS_MBYTE)) : INFINITY
 
 /* Format string to display a protocol performance function */
-#define UCP_PROTO_PERF_FUNC_FMT(_perf_var) " " #_perf_var ": " \
+#define UCP_PROTO_PERF_FUNC_FMT \
     UCP_PROTO_PERF_FUNC_TIME_FMT " ns/KB, " \
     UCP_PROTO_PERF_FUNC_BW_FMT " MB/s"
 #define UCP_PROTO_PERF_FUNC_ARG(_perf_func) \
     UCP_PROTO_PERF_FUNC_TIME_ARG(_perf_func), \
     UCP_PROTO_PERF_FUNC_BW_ARG(_perf_func)
 
-/* Format string to display a protocol performance estimations
- * of different types. See ucp_proto_perf_type_t */
-#define UCP_PROTO_PERF_FUNC_TYPES_FMT \
-    UCP_PROTO_PERF_FUNC_FMT(single) \
-    UCP_PROTO_PERF_FUNC_FMT(multi) \
-    UCP_PROTO_PERF_FUNC_FMT(cpu)
-#define UCP_PROTO_PERF_FUNC_TYPES_ARG(_perf_func) \
-    UCP_PROTO_PERF_FUNC_ARG((&(_perf_func)[UCP_PROTO_PERF_TYPE_SINGLE])), \
-    UCP_PROTO_PERF_FUNC_ARG((&(_perf_func)[UCP_PROTO_PERF_TYPE_MULTI])), \
-    UCP_PROTO_PERF_FUNC_ARG((&(_perf_func)[UCP_PROTO_PERF_TYPE_CPU]))
-
 
 /*
  * Protocol performance node types
@@ -50,7 +44,6 @@
 typedef enum {
     UCP_PROTO_PERF_NODE_TYPE_DATA,   /* Data node */
     UCP_PROTO_PERF_NODE_TYPE_SELECT, /* Select one of children */
-    UCP_PROTO_PERF_NODE_TYPE_COMPOSE /* Compose new value from the children */
 } ucp_proto_perf_node_type_t;
 
 
@@ -59,8 +52,9 @@ void ucp_proto_select_perf_str(const ucs_linear_func_t *perf, char *time_str,
                                size_t bw_str_max);
 
 
-void ucp_proto_select_init_trace_caps(
-        const ucp_proto_init_params_t *init_params);
+void ucp_proto_select_init_trace_perf(const ucp_proto_init_params_t *init_params,
+                                      const ucp_proto_perf_t *perf,
+                                      const void *priv);
 
 
 void ucp_proto_select_info(ucp_worker_h worker,
@@ -92,6 +86,11 @@ void ucp_proto_config_info_str(ucp_worker_h worker,
                                size_t msg_length, ucs_string_buffer_t *strb);
 
 
+ucp_proto_perf_node_t *
+ucp_proto_perf_node_new(ucp_proto_perf_node_type_t type,
+                        unsigned selected_child, const char *name,
+                        const char *desc_fmt, va_list ap);
+
 ucp_proto_perf_node_t *
 ucp_proto_perf_node_new_data(const char *name, const char *desc_fmt, ...);
 
@@ -101,16 +100,16 @@ ucp_proto_perf_node_new_select(const char *name, unsigned selected_child,
                                const char *desc_fmt, ...);
 
 
-ucp_proto_perf_node_t *
-ucp_proto_perf_node_new_compose(const char *name, const char *desc_fmt, ...);
-
-
 void ucp_proto_perf_node_ref(ucp_proto_perf_node_t *perf_node);
 
 
 void ucp_proto_perf_node_deref(ucp_proto_perf_node_t **perf_node_p);
 
 
+ucp_proto_perf_node_t *
+ucp_proto_perf_node_dup(const ucp_proto_perf_node_t *perf_node);
+
+
 void ucp_proto_perf_node_own_child(ucp_proto_perf_node_t *perf_node,
                                    ucp_proto_perf_node_t **child_perf_node_p);
 
@@ -129,6 +128,11 @@ void ucp_proto_perf_node_add_data(ucp_proto_perf_node_t *perf_node,
                                   const ucs_linear_func_t value);
 
 
+void ucp_proto_perf_node_update_data(ucp_proto_perf_node_t *perf_node,
+                                     const char *name,
+                                     const ucs_linear_func_t value);
+
+
 void ucp_proto_perf_node_add_scalar(ucp_proto_perf_node_t *perf_node,
                                     const char *name, double value);
 
@@ -155,4 +159,12 @@ void ucp_proto_select_elem_trace(ucp_worker_h worker,
                                  ucp_proto_select_elem_t *select_elem);
 
 
+void ucp_proto_select_write_info(
+        ucp_worker_h worker,
+        const ucp_proto_select_init_protocols_t *proto_init,
+        const ucs_dynamic_bitmap_t *proto_mask, unsigned selected_idx,
+        ucp_proto_config_t *selected_config, size_t range_start,
+        size_t range_end);
+
+
 #endif
