@@ -25,12 +25,13 @@ ucp_ep_flush_request_update_uct_comp(ucp_request_t *req, int diff,
                 &req->send.state.uct_comp, req->send.state.uct_comp.count,
                 diff);
     ucs_assertv(!(req->send.flush.started_lanes & new_started_lanes),
-                "req=%p started_lanes=0x%x new_started_lanes=0x%x", req,
-                req->send.flush.started_lanes, new_started_lanes);
+                "req=%p started_lanes=0x%" PRIx64
+                " new_started_lanes=0x%" PRIx64,
+                req, req->send.flush.started_lanes, new_started_lanes);
 
     ucp_trace_req(req,
                   "flush update ep %p comp_count %d->%d num_lanes %d->%d "
-                  "started_lanes 0x%x->0x%x",
+                  "started_lanes 0x%" PRIx64 "->0x%" PRIx64,
                   req->send.ep, req->send.state.uct_comp.count,
                   req->send.state.uct_comp.count + diff,
                   req->send.flush.num_lanes, ucp_ep_num_lanes(req->send.ep),
@@ -99,7 +100,8 @@ static void ucp_ep_flush_progress(ucp_request_t *req)
     }
 
     ucp_trace_req(req,
-                  "progress ep=%p flush flags=0x%x started_lanes=0x%x count=%d",
+                  "progress ep=%p flush flags=0x%x started_lanes=0x%" PRIx64
+                  " count=%d",
                   ep, ep->flags, req->send.flush.started_lanes,
                   req->send.state.uct_comp.count);
 
@@ -229,20 +231,22 @@ static void ucp_ep_flush_request_resched(ucp_ep_h ep, ucp_request_t *req)
             (ucp_ep_config(ep)->p2p_lanes &&
              ep->worker->context->config.ext.proto_request_reset)) {
             ucs_assertv(!req->send.flush.started_lanes,
-                        "req=%p flush started_lanes=0x%x", req,
+                        "req=%p flush started_lanes=0x%" PRIx64, req,
                         req->send.flush.started_lanes);
         } else {
-            ucs_assertv(!(UCS_BIT(req->send.lane) & req->send.flush.started_lanes),
-                        "req=%p lane=%d started_lanes=0x%x", req, req->send.lane,
-                        req->send.flush.started_lanes);
+            ucs_assertv(!(UCS_BIT(req->send.lane) &
+                          req->send.flush.started_lanes),
+                        "req=%p lane=%d started_lanes=0x%" PRIx64, req,
+                        req->send.lane, req->send.flush.started_lanes);
 
             /* Only lanes connected to iface can be started/flushed before
              * wireup is done because connect2iface does not create wireup_ep
              * without cm mode */
             ucs_assertv(!(req->send.flush.started_lanes &
                           ucp_ep_config(ep)->p2p_lanes),
-                        "req=%p flush started_lanes=0x%x p2p_lanes=0x%x", req,
-                        req->send.flush.started_lanes,
+                        "req=%p flush started_lanes=0x%" PRIx64
+                        " p2p_lanes=0x%" PRIx64,
+                        req, req->send.flush.started_lanes,
                         ucp_ep_config(ep)->p2p_lanes);
         }
 
@@ -348,9 +352,10 @@ void ucp_ep_flush_request_ff(ucp_request_t *req, ucs_status_t status)
     int num_comps = req->send.flush.num_lanes -
                     ucs_popcount(req->send.flush.started_lanes);
 
-    ucp_trace_req(req, "fast-forward flush, comp-=%d num_lanes %d started 0x%x",
-                  num_comps, req->send.flush.num_lanes,
-                  req->send.flush.started_lanes);
+    ucp_trace_req(
+            req, "fast-forward flush, comp-=%d num_lanes %d started 0x%" PRIx64,
+            num_comps, req->send.flush.num_lanes,
+            req->send.flush.started_lanes);
 
     ucp_ep_flush_request_update_uct_comp(req, -num_comps,
                                          UCS_MASK(req->send.flush.num_lanes) &
@@ -703,7 +708,7 @@ static ucs_status_t ucp_worker_fence_weak(ucp_worker_h worker)
     ucp_rsc_index_t rsc_index;
     ucs_status_t status;
 
-    UCS_BITMAP_FOR_EACH_BIT(worker->context->tl_bitmap, rsc_index) {
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(rsc_index, &worker->context->tl_bitmap) {
         wiface = ucp_worker_iface(worker, rsc_index);
         if (wiface->iface == NULL) {
             continue;
@@ -720,31 +725,10 @@ static ucs_status_t ucp_worker_fence_weak(ucp_worker_h worker)
 
 static ucs_status_t ucp_worker_fence_strong(ucp_worker_h worker)
 {
-    ucp_request_t *request;
-    ucs_status_ptr_t status_ptr;
-    ucs_status_t status;
-
-    ucp_worker_flush_ops_count_add(worker, 1);
-    status_ptr = ucp_worker_flush_nbx_internal(worker,
-                                                &ucp_request_null_param,
-                                                UCT_FLUSH_FLAG_REMOTE);
-    if (ucs_unlikely(!UCS_PTR_IS_PTR(status_ptr))) {
-        status = UCS_PTR_STATUS(status_ptr);
-        goto out;
-    }
-
-    request = (ucp_request_t*)status_ptr - 1;
-
-    do {
-        ucp_worker_progress(worker);
-    } while (request->flush_worker.comp_count > 1);
-
-    ucp_worker_flush_complete_one(request, request->status, 1);
-    status = request->status;
-    ucp_request_put(request);
-out:
-    ucp_worker_flush_ops_count_add(worker, -1);
-    return status;
+    void *request = ucp_worker_flush_nbx_internal(worker,
+                                                  &ucp_request_null_param,
+                                                  UCT_FLUSH_FLAG_REMOTE);
+    return ucp_rma_wait(worker, request, "strong_fence");
 }
 
 UCS_PROFILE_FUNC(ucs_status_t, ucp_worker_fence, (worker), ucp_worker_h worker)
@@ -753,14 +737,11 @@ UCS_PROFILE_FUNC(ucs_status_t, ucp_worker_fence, (worker), ucp_worker_h worker)
 
     UCP_WORKER_THREAD_CS_ENTER_CONDITIONAL(worker);
 
-    /* TODO: need to fix remote flush to ensure ordering in case of relaxed
-     * ordering, then calling weak fence is not needed in strong mode. */
-    status = ucp_worker_fence_weak(worker);
-
-    if (worker->context->config.worker_strong_fence &&
-        !UCS_STATUS_IS_ERR(status)) {
+    if (worker->context->config.worker_strong_fence) {
         /* force using flush on EPs */
         status = ucp_worker_fence_strong(worker);
+    } else {
+        status = ucp_worker_fence_weak(worker);
     }
 
     UCP_WORKER_THREAD_CS_EXIT_CONDITIONAL(worker);
