@@ -14,7 +14,7 @@
 #include <ucp/core/ucp_worker.h>
 #include <ucp/core/ucp_ep.inl>
 #include <ucs/arch/bitops.h>
-#include <ucs/datastruct/array.inl>
+#include <ucs/datastruct/array.h>
 #include <ucs/debug/log.h>
 #include <ucs/type/serialize.h>
 #include <ucs/type/float8.h>
@@ -172,10 +172,8 @@ typedef struct {
     ucs_sys_device_t sys_dev;
 } ucp_address_remote_device_t;
 
-UCS_ARRAY_DEFINE_INLINE(ucp_address_remote_device, unsigned,
-                        ucp_address_remote_device_t);
-
-typedef ucs_array_t(ucp_address_remote_device) ucp_address_remote_device_array_t;
+UCS_ARRAY_DECLARE_TYPE(ucp_address_remote_device_array_t, unsigned,
+                       ucp_address_remote_device_t);
 
 #define UCP_ADDRESS_V1_FLAG_ATOMIC32  UCS_BIT(30) /* 32bit atomic operations */
 #define UCP_ADDRESS_V1_FLAG_ATOMIC64  UCS_BIT(31) /* 64bit atomic operations */
@@ -322,13 +320,19 @@ out:
     return dev;
 }
 
-static size_t ucp_address_packed_value_size(size_t value, size_t max_value,
-                                            ucp_object_version_t addr_version)
+static ssize_t ucp_address_packed_value_size(size_t value, size_t max_value,
+                                             ucp_object_version_t addr_version,
+                                             const char *dev_name)
 {
     if (addr_version == UCP_OBJECT_VERSION_V1) {
         /* Address version 1 does not support value extension */
-        ucs_assertv_always(value <= max_value, "value %zu, max_value %zu",
-                           value, max_value);
+        if (value > max_value) {
+            ucs_debug("device %s: value %zu > max_value %zu, use "
+                      "UCX_ADDRESS_VERSION=v2 to allow packing such addresses",
+                      dev_name, value, max_value);
+            return UCS_ERR_UNSUPPORTED;
+        }
+
         return sizeof(uint8_t);
     } else if (value < max_value) {
         /* The value fits into a partial byte, up to max_value */
@@ -340,15 +344,17 @@ static size_t ucp_address_packed_value_size(size_t value, size_t max_value,
     }
 }
 
-static size_t ucp_address_packed_length_size(ucp_worker_h worker, size_t length,
-                                             size_t max_length,
-                                             ucp_object_version_t addr_version)
+static ssize_t ucp_address_packed_length_size(ucp_worker_h worker,
+                                              size_t length, size_t max_length,
+                                              ucp_object_version_t addr_version,
+                                              const char *dev_name)
 {
     if (ucp_worker_is_unified_mode(worker)) {
         return 0;
     }
 
-    return ucp_address_packed_value_size(length, max_length, addr_version);
+    return ucp_address_packed_value_size(length, max_length, addr_version,
+                                         dev_name);
 }
 
 static int ucp_address_pack_v1_extra_info(ucp_object_version_t addr_version,
@@ -373,6 +379,7 @@ ucp_address_gather_devices(ucp_worker_h worker, const ucp_ep_config_key_t *key,
     ucp_rsc_index_t num_devices;
     ucp_rsc_index_t rsc_index;
     ucp_lane_index_t lane;
+    ssize_t length_size;
 
     devices = ucs_calloc(context->num_tls, sizeof(*devices), "packed_devices");
     if (devices == NULL) {
@@ -380,8 +387,8 @@ ucp_address_gather_devices(ucp_worker_h worker, const ucp_ep_config_key_t *key,
     }
 
     num_devices = 0;
-    UCS_BITMAP_AND_INPLACE(&current_tl_bitmap, context->tl_bitmap);
-    UCS_BITMAP_FOR_EACH_BIT(current_tl_bitmap, rsc_index) {
+    UCS_STATIC_BITMAP_AND_INPLACE(&current_tl_bitmap, context->tl_bitmap);
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(rsc_index, &current_tl_bitmap) {
         iface_attr = ucp_worker_iface_get_attr(worker, rsc_index);
         if (!ucp_worker_iface_can_connect(iface_attr)) {
             continue;
@@ -412,9 +419,16 @@ ucp_address_gather_devices(ucp_worker_h worker, const ucp_ep_config_key_t *key,
             /* iface address length (+flags) can take 2 bytes with address
              * version 2 in non-unified mode
              */
-            dev->tl_addrs_size += ucp_address_packed_length_size(
-                                      worker, iface_attr->iface_addr_len,
-                                      UCP_ADDRESS_IFACE_LEN_MASK, addr_version);
+            length_size         = ucp_address_packed_length_size(
+                    worker, iface_attr->iface_addr_len,
+                    UCP_ADDRESS_IFACE_LEN_MASK, addr_version,
+                    worker->context->tl_rscs[rsc_index].tl_rsc.dev_name);
+            if (length_size < 0) {
+                ucs_free(devices);
+                return length_size;
+            }
+
+            dev->tl_addrs_size += length_size;
             dev->tl_addrs_size += ucp_address_iface_attr_size(worker, flags,
                                                               addr_version);
         } else {
@@ -441,7 +455,7 @@ ucp_address_gather_devices(ucp_worker_h worker, const ucp_ep_config_key_t *key,
         }
 
         dev->rsc_index  = rsc_index;
-        UCS_BITMAP_SET(dev->tl_bitmap, rsc_index);
+        UCS_STATIC_BITMAP_SET(&dev->tl_bitmap, rsc_index);
         dev->num_paths  = ucs_min(max_num_paths, iface_attr->dev_num_paths);
     }
 
@@ -450,7 +464,7 @@ ucp_address_gather_devices(ucp_worker_h worker, const ucp_ep_config_key_t *key,
     return UCS_OK;
 }
 
-static size_t
+static ssize_t
 ucp_address_packed_size(ucp_worker_h worker,
                         const ucp_address_packed_device_t *devices,
                         ucp_rsc_index_t num_devices, uint64_t pack_flags,
@@ -458,8 +472,10 @@ ucp_address_packed_size(ucp_worker_h worker,
 {
     size_t size = 0;
     size_t md_mask;
+    ssize_t value_size;
     const ucp_address_packed_device_t *dev;
     ucp_md_index_t md_index;
+    const ucp_tl_resource_desc_t *rsc;
 
     /* header: version and flags */
     if (addr_version == UCP_OBJECT_VERSION_V1) {
@@ -487,16 +503,28 @@ ucp_address_packed_size(ucp_worker_h worker,
         size += 1; /* NULL md_index */
     } else {
         for (dev = devices; dev < (devices + num_devices); ++dev) {
+            rsc        = &worker->context->tl_rscs[dev->rsc_index];
             /* device md_index */
-            md_index = worker->context->tl_rscs[dev->rsc_index].md_index;
+            md_index   = rsc->md_index;
             /* md index (+flags) can take 2 bytes with address version 2 */
-            size    += ucp_address_packed_value_size(md_index, md_mask,
-                                                     addr_version);
+            value_size = ucp_address_packed_value_size(md_index, md_mask,
+                                                       addr_version,
+                                                       rsc->tl_rsc.dev_name);
+            if (value_size < 0) {
+                return value_size;
+            }
+
+            size += value_size;
             if (pack_flags & UCP_ADDRESS_PACK_FLAG_DEVICE_ADDR) {
                 /* device address length */
-                size += ucp_address_packed_value_size(
-                            dev->dev_addr_len, UCP_ADDRESS_DEVICE_LEN_MASK,
-                            addr_version);
+                value_size = ucp_address_packed_value_size(
+                        dev->dev_addr_len, UCP_ADDRESS_DEVICE_LEN_MASK,
+                        addr_version, rsc->tl_rsc.dev_name);
+                if (value_size < 0) {
+                    return value_size;
+                }
+
+                size += value_size;
                 size += dev->dev_addr_len;  /* device address */
             } else {
                 size += 1; /* 0 device address length */
@@ -936,7 +964,7 @@ ucp_address_unpack_iface_attr(ucp_worker_t *worker,
         unified             = ptr;
         rsc_idx             = unified->rsc_index & UCP_ADDRESS_IFACE_LEN_MASK;
         iface_attr->lat_ovh = fabs(unified->lat_ovh);
-        if (!UCS_BITMAP_GET(worker->context->tl_bitmap, rsc_idx)) {
+        if (!UCS_STATIC_BITMAP_GET(worker->context->tl_bitmap, rsc_idx)) {
             ucp_address_error(
                     unpack_flags,
                     "failed to unpack address, resource[%d] is not valid",
@@ -954,7 +982,6 @@ ucp_address_unpack_iface_attr(ucp_worker_t *worker,
         iface_attr->bandwidth     =
                 ucp_tl_iface_bandwidth(worker->context,
                                        &wiface->attr.bandwidth);
-        iface_attr->dst_rsc_index = rsc_idx;
         iface_attr->seg_size      = wiface->attr.cap.am.max_bcopy;
 
         if (signbit(unified->lat_ovh)) {
@@ -998,11 +1025,9 @@ ucp_address_unpack_iface_attr(ucp_worker_t *worker,
     *size_p = iface_attr_len;
 
     if (unpack_flags & UCP_ADDRESS_PACK_FLAG_TL_RSC_IDX) {
-        ptr                       = UCS_PTR_BYTE_OFFSET(ptr, iface_attr_len);
-        iface_attr->dst_rsc_index = *(uint8_t*)ptr;
-        *size_p                  += sizeof(uint8_t);
-    } else {
-        iface_attr->dst_rsc_index = UCP_NULL_RESOURCE;
+        /* dst_rsc_index was removed from unpacked address, but still exists
+         * in packed address (for wire compatibility), so we need to skip it */
+        *size_p += sizeof(uint8_t);
     }
 
     return UCS_OK;
@@ -1261,15 +1286,15 @@ ucp_address_do_pack(ucp_worker_h worker, ucp_ep_h ep, void *buffer, size_t size,
 
     for (dev = devices; dev < (devices + num_devices); ++dev) {
         dev_tl_bitmap = context->tl_bitmap;
-        UCS_BITMAP_AND_INPLACE(&dev_tl_bitmap, dev->tl_bitmap);
+        UCS_STATIC_BITMAP_AND_INPLACE(&dev_tl_bitmap, dev->tl_bitmap);
 
         /* MD index */
         md_index      = context->tl_rscs[dev->rsc_index].md_index;
         md_flags      = context->tl_mds[md_index].attr.flags &
                             md_flags_pack_mask;
-        ptr           = ucp_address_pack_md_info(
-                            ptr, UCS_BITMAP_IS_ZERO_INPLACE(&dev_tl_bitmap),
-                            md_flags, md_index, addr_version);
+        ptr           = ucp_address_pack_md_info(ptr,
+                                                 UCS_STATIC_BITMAP_IS_ZERO(dev_tl_bitmap),
+                                                 md_flags, md_index, addr_version);
         dev_flags_ptr = ptr;
         ucs_assert_always((pack_flags & UCP_ADDRESS_PACK_FLAG_DEVICE_ADDR) ||
                           (dev->dev_addr_len == 0));
@@ -1311,7 +1336,7 @@ ucp_address_do_pack(ucp_worker_h worker, ucp_ep_h ep, void *buffer, size_t size,
         }
 
         flags_ptr = NULL;
-        UCS_BITMAP_FOR_EACH_BIT(dev_tl_bitmap, rsc_index) {
+        UCS_STATIC_BITMAP_FOR_EACH_BIT(rsc_index, &dev_tl_bitmap) {
             wiface     = ucp_worker_iface(worker, rsc_index);
             iface_attr = &wiface->attr;
 
@@ -1331,7 +1356,7 @@ ucp_address_do_pack(ucp_worker_h worker, ucp_ep_h ep, void *buffer, size_t size,
                                       context->tl_rscs[rsc_index].tl_name_csum);
 
             /* Transport information */
-            enable_amo = UCS_BITMAP_GET(worker->atomic_tls, rsc_index);
+            enable_amo = UCS_STATIC_BITMAP_GET(worker->atomic_tls, rsc_index);
             attr_len   = ucp_address_pack_iface_attr(wiface, ptr, rsc_index,
                                                      pack_flags, addr_version,
                                                      enable_amo);
@@ -1452,7 +1477,8 @@ ucp_address_do_pack(ucp_worker_h worker, ucp_ep_h ep, void *buffer, size_t size,
                               UCT_TL_RESOURCE_DESC_ARG(
                                       &context->tl_rscs[rsc_index].tl_rsc),
                               dev->sys_dev, dev->num_paths, num_ep_addrs,
-                              md_flags, iface_attr->cap.flags,
+                              md_flags,
+                              iface_attr->cap.flags & UCP_ADDRESS_IFACE_FLAGS,
                               iface_attr->bandwidth.dedicated / UCS_MBYTE,
                               iface_attr->bandwidth.shared / UCS_MBYTE,
                               iface_attr->overhead * 1e9,
@@ -1470,11 +1496,11 @@ ucp_address_do_pack(ucp_worker_h worker, ucp_ep_h ep, void *buffer, size_t size,
          * during the above loop So, set the LAST flag for the flags_ptr
          * from the last iteration */
         if (flags_ptr != NULL) {
-            ucs_assert(!UCS_BITMAP_IS_ZERO_INPLACE(&dev_tl_bitmap));
+            ucs_assert(!UCS_STATIC_BITMAP_IS_ZERO(dev_tl_bitmap));
             *(uint8_t*)flags_ptr |= UCP_ADDRESS_FLAG_LAST;
         } else {
             /* cppcheck-suppress internalAstError */
-            ucs_assert(UCS_BITMAP_IS_ZERO_INPLACE(&dev_tl_bitmap));
+            ucs_assert(UCS_STATIC_BITMAP_IS_ZERO(dev_tl_bitmap));
         }
     }
 
@@ -1509,6 +1535,7 @@ ucp_address_length(ucp_worker_h worker, const ucp_ep_config_key_t *key,
     ucp_address_packed_device_t *devices;
     ucp_rsc_index_t num_devices;
     ucs_status_t status;
+    ssize_t size;
 
     /* Collect all devices required to pack their address */
     status = ucp_address_gather_devices(worker, key, tl_bitmap, pack_flags,
@@ -1519,9 +1546,14 @@ ucp_address_length(ucp_worker_h worker, const ucp_ep_config_key_t *key,
     }
 
     /* Calculate the required ucp address length */
-    *size_p = ucp_address_packed_size(worker, devices, num_devices, pack_flags,
-                                      addr_version);
-    status  = UCS_OK;
+    size = ucp_address_packed_size(worker, devices, num_devices, pack_flags,
+                                   addr_version);
+    if (size > -1) {
+        *size_p = size;
+    } else {
+        status = size;
+    }
+
     ucs_free(devices);
 
 out:
@@ -1541,7 +1573,7 @@ ucs_status_t ucp_address_pack(ucp_worker_h worker, ucp_ep_h ep,
     const ucp_ep_config_key_t *key;
     ucs_status_t status;
     void *buffer;
-    size_t size;
+    ssize_t size;
 
     if (ep == NULL) {
         pack_flags &= ~UCP_ADDRESS_PACK_FLAG_EP_ADDR;
@@ -1561,6 +1593,10 @@ ucs_status_t ucp_address_pack(ucp_worker_h worker, ucp_ep_h ep,
     /* Calculate packed size */
     size = ucp_address_packed_size(worker, devices, num_devices, pack_flags,
                                    addr_version);
+    if (size < 0) {
+        status = (ucs_status_t)size;
+        goto out_free_devices;
+    }
 
     /* Allocate address */
     buffer = ucs_malloc(size, "ucp_address");
@@ -1611,8 +1647,7 @@ static ucp_rsc_index_t ucp_address_get_remote_device_index(
         unique_dev_index++;
     }
 
-    device            = ucs_array_append_fixed(ucp_address_remote_device,
-                                               device_array);
+    device            = ucs_array_append_fixed(device_array);
     device->sys_dev   = sys_dev;
     device->dev_index = dev_index;
     return ucs_array_length(device_array) - 1;
@@ -1622,8 +1657,8 @@ ucs_status_t ucp_address_unpack(ucp_worker_t *worker, const void *buffer,
                                 unsigned unpack_flags,
                                 ucp_unpacked_address_t *unpacked_address)
 {
-    UCS_ARRAY_DEFINE_ONSTACK(remote_device_array, ucp_address_remote_device,
-                             UCP_MAX_RESOURCES);
+    UCS_ARRAY_DEFINE_ONSTACK(ucp_address_remote_device_array_t,
+                             remote_device_array, UCP_MAX_RESOURCES);
     ucp_address_entry_t *address_list, *address;
     uint8_t addr_flags;
     ucp_object_version_t addr_version;
@@ -1816,7 +1851,7 @@ ucs_status_t ucp_address_unpack(ucp_worker_t *worker, const void *buffer,
 
             ucp_address_trace(unpack_flags,
                               "unpack addr[%d] : sysdev %d paths %d eps %u"
-                              " tl_iface_flags 0x%" PRIx64 " bw %.2f/nMBs"
+                              " tl_flags 0x%" PRIx64 " bw %.2f/nMBs"
                               " ovh %.0fns lat_ovh %.0fns dev_priority %d"
                               " a32 0x%" PRIx64 "/0x%" PRIx64 " a64 0x%" PRIx64
                               "/0x%" PRIx64,
