@@ -328,7 +328,7 @@ ucs_status_t ucs_socket_sendv_nb(int fd, struct iovec *iov, size_t iov_cnt,
  * @param [in]      data            A pointer to a buffer to receive the incoming
  *                                  data.
  * @param [in/out]  length          The length, in bytes, of the data in buffer
- *                                  pointed to by the `data` paramete.
+ *                                  pointed to by the `data` parameter.
  *
  * @return UCS_OK on success or an error code on failure.
  */
@@ -539,6 +539,21 @@ int ucs_sockaddr_cmp(const struct sockaddr *sa1,
 int ucs_sockaddr_ip_cmp(const struct sockaddr *sa1, const struct sockaddr *sa2);
 
 
+/**
+ * Checks if two IP addresses are in the same subnet
+ *
+ * @param [in]  sa1         Pointer to sockaddr structure #1.
+ * @param [in]  sa2         Pointer to sockaddr structure #2.
+ * @param [in]  prefix_len  Subnet prefix length in bits.
+ *
+ * @return 1 if both addresses are in the same subnet.
+ *         0 if not
+ */
+int ucs_sockaddr_is_same_subnet(const struct sockaddr *sa1,
+                                const struct sockaddr *sa2,
+                                unsigned prefix_len);
+
+
 /**
  * Indicate if given IP address is INADDR_ANY (IPV4) or in6addr_any (IPV6)
  *
