@@ -242,6 +242,26 @@ static int uct_iface_is_same_device(const uct_iface_h iface,
     return !memcmp(device_addr, dev_addr, attr.device_addr_len);
 }
 
+void uct_iface_fill_info_str_buf(const uct_iface_is_reachable_params_t *params,
+                                 const char *fmt, ...)
+{
+    char *info_str_buf      = UCS_PARAM_VALUE(UCT_IFACE_IS_REACHABLE_FIELD,
+                                              params, info_string, INFO_STRING,
+                                              NULL);
+    size_t info_str_buf_len = UCS_PARAM_VALUE(UCT_IFACE_IS_REACHABLE_FIELD,
+                                              params, info_string_length,
+                                              INFO_STRING_LENGTH, 0);
+    va_list ap;
+
+    if (info_str_buf == NULL) {
+        return;
+    }
+
+    va_start(ap, fmt);
+    ucs_vsnprintf_safe(info_str_buf, info_str_buf_len, fmt, ap);
+    va_end(ap);
+}
+
 int uct_iface_is_reachable_params_valid(
         const uct_iface_is_reachable_params_t *params, uint64_t flags)
 {
@@ -281,14 +301,33 @@ int uct_iface_scope_is_reachable(const uct_iface_h iface,
 
     ucs_assert(params->field_mask & UCT_IFACE_IS_REACHABLE_FIELD_DEVICE_ADDR);
 
-    return (scope == UCT_IFACE_REACHABILITY_SCOPE_NETWORK) ||
-           uct_iface_is_same_device(iface, params->device_addr);
+    if (scope == UCT_IFACE_REACHABILITY_SCOPE_NETWORK) {
+        return 1;
+    }
+
+    if (!uct_iface_is_same_device(iface, params->device_addr)) {
+        uct_iface_fill_info_str_buf(params, "device address is different");
+        return 0;
+    }
+
+    return 1;
 }
 
 int uct_iface_is_reachable_v2(const uct_iface_h tl_iface,
                               const uct_iface_is_reachable_params_t *params)
 {
     const uct_base_iface_t *iface = ucs_derived_of(tl_iface, uct_base_iface_t);
+    char *info_str_buf            = UCS_PARAM_VALUE(
+                                          UCT_IFACE_IS_REACHABLE_FIELD, params,
+                                          info_string, INFO_STRING, NULL);
+    size_t info_str_buf_len       = UCS_PARAM_VALUE(
+                                          UCT_IFACE_IS_REACHABLE_FIELD, params,
+                                          info_string_length,
+                                          INFO_STRING_LENGTH, 0);
+
+    if ((info_str_buf != NULL) && (info_str_buf_len > 0)) {
+        info_str_buf[0] = '\0';
+    }
 
     return iface->internal_ops->iface_is_reachable_v2(tl_iface, params);
 }
@@ -310,9 +349,21 @@ int uct_base_iface_is_reachable(const uct_iface_h tl_iface,
 int uct_base_ep_is_connected(const uct_ep_h tl_ep,
                              const uct_ep_is_connected_params_t *params)
 {
-    UCT_EP_IS_CONNECTED_CHECK_DEV_IFACE_ADDRS(params);
-    return uct_base_iface_is_reachable(tl_ep->iface, params->device_addr,
-                                       params->iface_addr);
+    uct_iface_is_reachable_params_t is_reachable_params = {0};
+
+    if (params->field_mask & UCT_EP_IS_CONNECTED_FIELD_DEVICE_ADDR) {
+        is_reachable_params.field_mask |=
+                UCT_IFACE_IS_REACHABLE_FIELD_DEVICE_ADDR;
+        is_reachable_params.device_addr = params->device_addr;
+    }
+
+    if (params->field_mask & UCT_EP_IS_CONNECTED_FIELD_IFACE_ADDR) {
+        is_reachable_params.field_mask |=
+                UCT_IFACE_IS_REACHABLE_FIELD_IFACE_ADDR;
+        is_reachable_params.iface_addr  = params->iface_addr;
+    }
+
+    return uct_iface_is_reachable_v2(tl_ep->iface, &is_reachable_params);
 }
 
 int uct_ep_is_connected(uct_ep_h ep, const uct_ep_is_connected_params_t *params)
@@ -949,7 +1000,8 @@ void uct_iface_get_local_address(uct_iface_local_addr_ns_t *addr_ns,
 }
 
 int uct_iface_local_is_reachable(uct_iface_local_addr_ns_t *addr_ns,
-                                 ucs_sys_namespace_type_t sys_ns_type)
+                                 ucs_sys_namespace_type_t sys_ns_type,
+                                 const uct_iface_is_reachable_params_t *params)
 {
     uct_iface_local_addr_ns_t my_addr = {};
 
@@ -960,6 +1012,9 @@ int uct_iface_local_is_reachable(uct_iface_local_addr_ns_t *addr_ns,
     /* Check if both processes are on same host and both of them are in root (or
      * non-root) pid namespace */
     if (addr_ns->super.id != my_addr.super.id) {
+        uct_iface_fill_info_str_buf(params,
+                                    "different host id local %lu vs %lu",
+                                    my_addr.super.id, addr_ns->super.id);
         return 0;
     }
 
@@ -969,7 +1024,14 @@ int uct_iface_local_is_reachable(uct_iface_local_addr_ns_t *addr_ns,
 
     /* We are in non-root PID namespace - return 1 if ID of namespaces are the
      * same */
-    return addr_ns->sys_ns == my_addr.sys_ns;
+    if (addr_ns->sys_ns != my_addr.sys_ns) {
+        uct_iface_fill_info_str_buf(
+                    params,
+                    "different pid namespaces %"PRIx64" vs %"PRIx64"",
+                    my_addr.sys_ns, addr_ns->sys_ns);
+        return 0;
+    }
+    return 1;
 }
 
 void uct_iface_mpool_config_copy(ucs_mpool_params_t *mp_params,
