@@ -67,19 +67,19 @@ ucs_config_field_t uct_mm_iface_config_table[] = {
     {"ERROR_HANDLING", "n", "Expose error handling support capability",
      ucs_offsetof(uct_mm_iface_config_t, error_handling), UCS_CONFIG_TYPE_BOOL},
 
-    {"SEND_OVERHEAD", UCS_VALUE_AUTO_STR,
+    {"SEND_OVERHEAD", UCS_PP_MAKE_STRING(UCT_MM_IFACE_OVERHEAD),
      "Time spent after the message request has been passed to the hardware or\n"
      "system software layers and before operation has been finalized", 0,
-     UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME_UNITS,
+     UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME,
         {"am_short", "send overhead for short Active Message operation type",
          ucs_offsetof(uct_mm_iface_config_t, overhead.send.am_short)},
         {"am_bcopy", "send overhead for buffered Active Message operation type",
          ucs_offsetof(uct_mm_iface_config_t, overhead.send.am_bcopy)},
         {NULL})},
 
-    {"RECV_OVERHEAD", UCS_VALUE_AUTO_STR,
+    {"RECV_OVERHEAD", UCS_PP_MAKE_STRING(UCT_MM_IFACE_OVERHEAD),
      "Message receive overhead time", 0,
-     UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME_UNITS,
+     UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME,
         {"am_short", "receive overhead for short Active Message operation type",
          ucs_offsetof(uct_mm_iface_config_t, overhead.recv.am_short)},
         {"am_bcopy", "receive overhead for buffered Active Message operation "
@@ -138,8 +138,12 @@ uct_mm_iface_is_reachable_v2(const uct_iface_h tl_iface,
     }
 
     iface_addr = (void*)params->iface_addr;
+    if (iface_addr == NULL) {
+        uct_iface_fill_info_str_buf(params, "iface address is empty");
+        return 0;
+    }
 
-    return uct_sm_iface_is_reachable(tl_iface, params->device_addr) &&
+    return uct_sm_iface_is_reachable(tl_iface, params) &&
            uct_mm_md_mapper_ops(md)->is_reachable(md, iface_addr->fifo_seg_id,
                                                   iface_addr + 1) &&
            uct_iface_scope_is_reachable(tl_iface, params);
@@ -537,7 +541,6 @@ uct_mm_estimate_perf(uct_iface_h tl_iface, uct_perf_attr_t *perf_attr)
     uct_mm_iface_t *iface = ucs_derived_of(tl_iface, uct_mm_iface_t);
     uct_ep_operation_t op = UCT_ATTR_VALUE(PERF, perf_attr, operation,
                                            OPERATION, UCT_EP_OP_LAST);
-    double short_overhead, am_overhead;
     uct_mm_iface_op_overhead_t *overhead;
 
     if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_BANDWIDTH) {
@@ -545,26 +548,14 @@ uct_mm_estimate_perf(uct_iface_h tl_iface, uct_perf_attr_t *perf_attr)
         perf_attr->bandwidth.dedicated = iface->super.config.bandwidth;
     }
 
-    switch (ucs_arch_get_cpu_vendor()) {
-    case UCS_CPU_VENDOR_FUJITSU_ARM:
-        short_overhead = 40e-9;
-        am_overhead    = 220e-9;
-        break;
-    default:
-        short_overhead = UCT_MM_IFACE_OVERHEAD;
-        am_overhead    = UCT_MM_IFACE_OVERHEAD;
-    }
-
     if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_SEND_PRE_OVERHEAD) {
         overhead = &iface->config.overhead.send;
         switch (op) {
         case UCT_EP_OP_AM_SHORT:
-            perf_attr->send_pre_overhead =
-                    ucs_time_units_to_sec(overhead->am_short, short_overhead);
+            perf_attr->send_pre_overhead = overhead->am_short;
             break;
         case UCT_EP_OP_AM_BCOPY:
-            perf_attr->send_pre_overhead =
-                    ucs_time_units_to_sec(overhead->am_bcopy, am_overhead);
+            perf_attr->send_pre_overhead = overhead->am_bcopy;
             break;
         default:
             perf_attr->send_pre_overhead = UCT_MM_IFACE_OVERHEAD;
@@ -576,12 +567,10 @@ uct_mm_estimate_perf(uct_iface_h tl_iface, uct_perf_attr_t *perf_attr)
         overhead = &iface->config.overhead.recv;
         switch (op) {
         case UCT_EP_OP_AM_SHORT:
-            perf_attr->recv_overhead = ucs_time_units_to_sec(overhead->am_short,
-                                                             short_overhead);
+            perf_attr->recv_overhead = overhead->am_short;
             break;
         case UCT_EP_OP_AM_BCOPY:
-            perf_attr->recv_overhead = ucs_time_units_to_sec(overhead->am_bcopy,
-                                                             am_overhead);
+            perf_attr->recv_overhead = overhead->am_bcopy;
             break;
         default:
             perf_attr->recv_overhead = UCT_MM_IFACE_OVERHEAD;
