@@ -111,6 +111,12 @@ static ucs_config_field_t uct_tcp_iface_config_table[] = {
                 UCS_CONFIG_TYPE_TIME_UNITS},
 #endif /* UCT_TCP_EP_KEEPALIVE */
 
+  {"EP_BIND_SRC_ADDR", "try",
+   "Bind client socket to the local network interface before connecting to the "
+   "remote peer",
+   ucs_offsetof(uct_tcp_iface_config_t, ep_bind_src_addr),
+                UCS_CONFIG_TYPE_TERNARY},
+
   {NULL}
 };
 
@@ -194,6 +200,7 @@ uct_tcp_iface_is_reachable_v2(const uct_iface_h tl_iface,
     uct_tcp_iface_t *iface = ucs_derived_of(tl_iface, uct_tcp_iface_t);
     uct_iface_local_addr_ns_t *local_addr_ns;
     uct_tcp_device_addr_t *tcp_dev_addr;
+    int is_local_loopback, is_remote_loopback;
 
     if (!uct_iface_is_reachable_params_valid(
                 params, UCT_IFACE_IS_REACHABLE_FIELD_DEVICE_ADDR)) {
@@ -202,19 +209,29 @@ uct_tcp_iface_is_reachable_v2(const uct_iface_h tl_iface,
 
     tcp_dev_addr = (uct_tcp_device_addr_t*)params->device_addr;
     if (iface->config.ifaddr.ss_family != tcp_dev_addr->sa_family) {
+        uct_iface_fill_info_str_buf(
+                params, "different address family %d vs %d",
+                iface->config.ifaddr.ss_family, tcp_dev_addr->sa_family);
         return 0;
     }
 
     /* Loopback can connect only to loopback */
-    if (!!(tcp_dev_addr->flags & UCT_TCP_DEVICE_ADDR_FLAG_LOOPBACK) !=
-        ucs_sockaddr_is_inaddr_loopback(
-                (const struct sockaddr*)&iface->config.ifaddr)) {
+    is_remote_loopback = !!(tcp_dev_addr->flags &
+                            UCT_TCP_DEVICE_ADDR_FLAG_LOOPBACK);
+    is_local_loopback  = ucs_sockaddr_is_inaddr_loopback(
+            (const struct sockaddr*)&iface->config.ifaddr);
+    if (is_remote_loopback != is_local_loopback) {
+        uct_iface_fill_info_str_buf(params,
+                                    "incompatible loopback flags, "
+                                    "%d (local) vs %d (remote)",
+                                    is_local_loopback, is_remote_loopback);
         return 0;
     }
 
-    if (tcp_dev_addr->flags & UCT_TCP_DEVICE_ADDR_FLAG_LOOPBACK) {
+    if (is_remote_loopback) {
         local_addr_ns = (uct_iface_local_addr_ns_t*)(tcp_dev_addr + 1);
-        if (!uct_iface_local_is_reachable(local_addr_ns, UCS_SYS_NS_TYPE_NET)) {
+        if (!uct_iface_local_is_reachable(local_addr_ns, UCS_SYS_NS_TYPE_NET,
+                                          params)) {
             return 0;
         }
     }
@@ -275,14 +292,15 @@ static ucs_status_t uct_tcp_iface_query(uct_iface_h tl_iface,
     attr->ep_addr_len      = sizeof(uct_tcp_ep_addr_t);
     attr->iface_addr_len   = sizeof(uct_tcp_iface_addr_t);
     attr->device_addr_len  = uct_tcp_iface_get_device_address_length(iface);
-    attr->cap.flags        = UCT_IFACE_FLAG_CONNECT_TO_IFACE |
-                             UCT_IFACE_FLAG_CONNECT_TO_EP    |
-                             UCT_IFACE_FLAG_AM_SHORT         |
-                             UCT_IFACE_FLAG_AM_BCOPY         |
-                             UCT_IFACE_FLAG_PENDING          |
-                             UCT_IFACE_FLAG_CB_SYNC          |
-                             UCT_IFACE_FLAG_EP_CHECK         |
-                             UCT_IFACE_FLAG_ERRHANDLE_PEER_FAILURE;
+    attr->cap.flags        = UCT_IFACE_FLAG_CONNECT_TO_IFACE       |
+                             UCT_IFACE_FLAG_CONNECT_TO_EP          |
+                             UCT_IFACE_FLAG_AM_SHORT               |
+                             UCT_IFACE_FLAG_AM_BCOPY               |
+                             UCT_IFACE_FLAG_PENDING                |
+                             UCT_IFACE_FLAG_CB_SYNC                |
+                             UCT_IFACE_FLAG_EP_CHECK               |
+                             UCT_IFACE_FLAG_ERRHANDLE_PEER_FAILURE |
+                             UCT_IFACE_FLAG_INTER_NODE;
     attr->cap.event_flags  = UCT_IFACE_FLAG_EVENT_SEND_COMP |
                              UCT_IFACE_FLAG_EVENT_RECV      |
                              UCT_IFACE_FLAG_EVENT_FD;
@@ -690,6 +708,7 @@ static UCS_CLASS_INIT_FUNC(uct_tcp_iface_t, uct_md_h md, uct_worker_h worker,
     self->sockopt.rcvbuf           = config->sockopt.rcvbuf;
     self->config.keepalive.cnt     = config->keepalive.cnt;
     self->config.keepalive.intvl   = config->keepalive.intvl;
+    self->config.ep_bind_src_addr  = config->ep_bind_src_addr;
     self->port_range.first         = config->port_range.first;
     self->port_range.last          = config->port_range.last;
 
