@@ -18,6 +18,7 @@ extern "C" {
 }
 
 #include <cmath>
+#include <list>
 
 class test_ucp_mmap : public ucp_test {
 public:
@@ -33,6 +34,16 @@ public:
         VARIANT_NO_RCACHE
     };
 
+    struct mem_chunk {
+        ucp_context_h           context;
+        ucp_mem_h               memh;
+        std::vector<ucp_rkey_h> rkeys;
+
+        mem_chunk(ucp_context_h);
+        ~mem_chunk();
+        ucp_rkey_h unpack(ucp_ep_h, ucp_md_map_t md_map = 0);
+    };
+
     static void
     get_test_variants(std::vector<ucp_test_variant>& variants,
                       uint64_t extra_features)
@@ -95,6 +106,11 @@ public:
             modify_config("PROTO_ENABLE", "n");
         }
 
+        if (get_variant_value() == VARIANT_MAP_NONBLOCK) {
+            // ODPv1 cannot interact with DEVX objects
+            modify_config("IB_MLX5_DEVX_OBJECTS", "", SETENV_IF_NOT_EXIST);
+        }
+
         if (get_variant_value() == VARIANT_NO_RCACHE) {
             modify_config("RCACHE_ENABLE", "n");
             ucp_test::init(); // Init UCP with rcache disabled
@@ -157,14 +173,15 @@ public:
         compare_uct_memhs(memh1, memh2);
     }
 
+    ucp_mem_h import_memh(ucp_mem_h exported_memh);
+
 protected:
     bool resolve_rma(entity *e, ucp_rkey_h rkey);
     bool resolve_amo(entity *e, ucp_rkey_h rkey);
     bool resolve_rma_bw_get_zcopy(entity *e, ucp_rkey_h rkey);
     bool resolve_rma_bw_put_zcopy(entity *e, ucp_rkey_h rkey);
     void test_length0(unsigned flags);
-    void test_rereg(unsigned map_flags = 0, uint64_t memh_pack_flags = 0,
-                    bool import_mem = false);
+    void test_rereg(unsigned map_flags = 0, bool import_mem = false);
     void test_rkey_management(ucp_mem_h memh, bool is_dummy,
                               bool expect_rma_offload);
 
@@ -179,15 +196,56 @@ private:
                                const char *function, ucs_log_level_t level,
                                const ucs_log_component_config_t *comp_conf,
                                const char *message, va_list ap);
-    void import_memh(void *exported_memh_buf, ucp_mem_h *memh_p);
     void release_exported_memh_buf(void *exported_memh_buf);
-    void test_rereg_imported_mem(ucp_mem_h memh, uint64_t memh_pack_flags,
-                                 size_t size);
+    void test_rereg_imported_mem(ucp_mem_h memh, size_t size);
 
 protected:
     ucp_md_map_t m_always_equal_md_map;
 };
 
+test_ucp_mmap::mem_chunk::mem_chunk(ucp_context_h ctx) : context(ctx)
+{
+    ucp_mem_map_params_t params = {
+        .field_mask = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
+                      UCP_MEM_MAP_PARAM_FIELD_LENGTH  |
+                      UCP_MEM_MAP_PARAM_FIELD_FLAGS,
+        .address    = NULL,
+        .length     = 4096,
+        .flags      = UCP_MEM_MAP_ALLOCATE,
+    };
+
+    ASSERT_UCS_OK(ucp_mem_map(context, &params, &memh));
+}
+
+test_ucp_mmap::mem_chunk::~mem_chunk()
+{
+    for (auto &rkey : rkeys) {
+        ucp_rkey_destroy(rkey);
+    }
+
+    EXPECT_UCS_OK(ucp_mem_unmap(context, memh));
+}
+
+ucp_rkey_h test_ucp_mmap::mem_chunk::unpack(ucp_ep_h ep, ucp_md_map_t md_map)
+{
+    ucp_rkey_h rkey;
+    void *rkey_buffer;
+    size_t rkey_size;
+
+    ASSERT_UCS_OK(ucp_rkey_pack(context, memh, &rkey_buffer, &rkey_size));
+    if (md_map == 0) {
+        ASSERT_UCS_OK(ucp_ep_rkey_unpack(ep, rkey_buffer, &rkey));
+    } else {
+        // Different MD map means different config index on proto v2
+        ASSERT_UCS_OK(ucp_ep_rkey_unpack_internal(ep, rkey_buffer, rkey_size,
+                                                  md_map, 0, &rkey));
+    }
+
+    ucp_rkey_buffer_release(rkey_buffer);
+    rkeys.push_back(rkey);
+    return rkey;
+}
+
 bool test_ucp_mmap::resolve_rma(entity *e, ucp_rkey_h rkey)
 {
     ucs_status_t status;
@@ -368,6 +426,9 @@ void test_ucp_mmap::check_distance_precision(double rkey_value,
     } else if (rkey_value == pack_max) {
         /* Capped by pack_max, no cache entry */
         EXPECT_GE(std::lround(topo_value), pack_max);
+    } else if (topo_value == INFINITY) {
+        /* Infinity values can be packed without loss */
+        EXPECT_EQ(topo_value, rkey_value);
     } else {
         /* Inside the borders or cache entry */
         EXPECT_NEAR(rkey_value, topo_value, topo_value * allowed_diff_ratio);
@@ -473,9 +534,9 @@ UCS_TEST_P(test_ucp_mmap, alloc_mem_type) {
             ASSERT_UCS_OK(status);
 
             is_dummy           = (size == 0);
-            expect_rma_offload = !UCP_MEM_IS_CUDA_MANAGED(mem_type) &&
-                                 (is_tl_rdma() || is_tl_shm()) &&
+            expect_rma_offload = (is_tl_rdma() || is_tl_shm()) &&
                                  check_reg_mem_types(sender(), mem_type);
+
             test_rkey_management(memh, is_dummy, expect_rma_offload);
 
             status = ucp_mem_unmap(sender().ucph(), memh);
@@ -494,7 +555,11 @@ UCS_TEST_P(test_ucp_mmap, reg_mem_type) {
 
     for (int i = 0; i < 1000 / ucs::test_time_multiplier(); ++i) {
         size_t size    = ucs::rand() % UCS_MBYTE;
-        alloc_mem_type = mem_types.at(ucs::rand() % mem_types.size());
+        auto flags     = mem_map_flags();
+        /* Test ODP registration with host buffer only */
+        alloc_mem_type = (flags & VARIANT_MAP_NONBLOCK) ?
+                UCS_MEMORY_TYPE_HOST :
+                mem_types.at(ucs::rand() % mem_types.size());
         mem_buffer buf(size, alloc_mem_type);
         mem_buffer::pattern_fill(buf.ptr(), size, 0, alloc_mem_type);
 
@@ -508,7 +573,7 @@ UCS_TEST_P(test_ucp_mmap, reg_mem_type) {
         params.address     = buf.ptr();
         params.length      = size;
         params.memory_type = alloc_mem_type;
-        params.flags       = mem_map_flags();
+        params.flags       = flags;
 
         status = ucp_mem_map(sender().ucph(), &params, &memh);
         ASSERT_UCS_OK(status);
@@ -518,8 +583,7 @@ UCS_TEST_P(test_ucp_mmap, reg_mem_type) {
             EXPECT_EQ(alloc_mem_type, memh->mem_type);
         }
 
-        expect_rma_offload = !UCP_MEM_IS_CUDA_MANAGED(alloc_mem_type) &&
-                             !UCP_MEM_IS_ROCM_MANAGED(alloc_mem_type) &&
+        expect_rma_offload = !UCP_MEM_IS_ROCM_MANAGED(alloc_mem_type) &&
                              is_tl_rdma() &&
                              check_reg_mem_types(sender(), alloc_mem_type);
         test_rkey_management(memh, is_dummy, expect_rma_offload);
@@ -592,28 +656,7 @@ void test_ucp_mmap::release_exported_memh_buf(void *exported_memh_buf)
     ucp_memh_buffer_release(exported_memh_buf, &release_params);
 }
 
-void test_ucp_mmap::import_memh(void *exported_memh_buf, ucp_mem_h *memh_p)
-{
-    ucp_mem_map_params_t params;
-
-    params.field_mask           =
-            UCP_MEM_MAP_PARAM_FIELD_EXPORTED_MEMH_BUFFER;
-    params.exported_memh_buffer = exported_memh_buf;
-
-    {
-        scoped_log_handler warn_slh(import_no_md_error_handler);
-        ucs_status_t status = ucp_mem_map(receiver().ucph(), &params, memh_p);
-        if (status == UCS_ERR_UNREACHABLE) {
-            release_exported_memh_buf(exported_memh_buf);
-            UCS_TEST_SKIP_R("memory importing is unsupported");
-        }
-        ASSERT_UCS_OK(status);
-    }
-}
-
-void test_ucp_mmap::test_rereg_imported_mem(ucp_mem_h memh,
-                                            uint64_t memh_pack_flags,
-                                            size_t size)
+ucp_mem_h test_ucp_mmap::import_memh(ucp_mem_h exported_memh)
 {
     ucp_memh_pack_params_t pack_params;
     ucs_status_t status;
@@ -621,23 +664,37 @@ void test_ucp_mmap::test_rereg_imported_mem(ucp_mem_h memh,
     size_t exported_memh_buf_size;
 
     pack_params.field_mask = UCP_MEMH_PACK_PARAM_FIELD_FLAGS;
-    pack_params.flags      = memh_pack_flags;
+    pack_params.flags      = UCP_MEMH_PACK_FLAG_EXPORT;
 
-    status = ucp_memh_pack(memh, &pack_params, &exported_memh_buf,
+    status = ucp_memh_pack(exported_memh, &pack_params, &exported_memh_buf,
                            &exported_memh_buf_size);
-    if ((status == UCS_ERR_UNSUPPORTED) &&
-        (pack_params.flags & UCP_MEMH_PACK_FLAG_EXPORT)) {
+    if (status == UCS_ERR_UNSUPPORTED) {
         UCS_TEST_SKIP_R("memory exporting is unsupported");
     }
     ASSERT_UCS_OK(status);
 
-    ucp_mem_h imp_memh;
-    import_memh(exported_memh_buf, &imp_memh);
+    ucp_mem_map_params_t params;
+    params.field_mask           =
+            UCP_MEM_MAP_PARAM_FIELD_EXPORTED_MEMH_BUFFER;
+    params.exported_memh_buffer = exported_memh_buf;
 
-    ucp_mem_h test_imp_memh;
-    import_memh(exported_memh_buf, &test_imp_memh);
+    ucp_mem_h imported_memh;
+    scoped_log_handler warn_slh(import_no_md_error_handler);
+    status = ucp_mem_map(receiver().ucph(), &params, &imported_memh);
+    if (status == UCS_ERR_UNREACHABLE) {
+        release_exported_memh_buf(exported_memh_buf);
+        UCS_TEST_SKIP_R("memory importing is unsupported");
+    }
 
     release_exported_memh_buf(exported_memh_buf);
+    ASSERT_UCS_OK(status);
+    return imported_memh;
+}
+
+void test_ucp_mmap::test_rereg_imported_mem(ucp_mem_h memh, size_t size)
+{
+    ucp_mem_h imp_memh      = import_memh(memh);
+    ucp_mem_h test_imp_memh = import_memh(memh);
 
     if (size == 0) {
         EXPECT_EQ(memh, test_imp_memh);
@@ -649,16 +706,11 @@ void test_ucp_mmap::test_rereg_imported_mem(ucp_mem_h memh,
         compare_memhs(test_imp_memh, imp_memh);
     }
 
-    status = ucp_mem_unmap(receiver().ucph(), test_imp_memh);
-    ASSERT_UCS_OK(status);
-
-    status = ucp_mem_unmap(receiver().ucph(), imp_memh);
-    ASSERT_UCS_OK(status);
+    ASSERT_UCS_OK(ucp_mem_unmap(receiver().ucph(), test_imp_memh));
+    ASSERT_UCS_OK(ucp_mem_unmap(receiver().ucph(), imp_memh));
 }
 
-void test_ucp_mmap::test_rereg(unsigned map_flags,
-                               uint64_t memh_pack_flags,
-                               bool import_mem)
+void test_ucp_mmap::test_rereg(unsigned map_flags, bool import_mem)
 {
     ucs_status_t status;
 
@@ -695,7 +747,7 @@ void test_ucp_mmap::test_rereg(unsigned map_flags,
 
         if (import_mem) {
             try {
-                test_rereg_imported_mem(memh, memh_pack_flags, size);
+                test_rereg_imported_mem(memh, size);
             } catch (ucs::test_skip_exception &e) {
                 status = ucp_mem_unmap(sender().ucph(), memh);
                 ASSERT_UCS_OK(status);
@@ -867,7 +919,7 @@ UCS_TEST_P(test_ucp_mmap, fixed) {
     /* Make sure eps are connected, because UCX async thread may add some
      * progress callbacks to worker callback queue
      * (e.g. ucp_worker_iface_check_events_progress) and mmap some memory
-     * for it (see ucs_callbackq_array_grow->ucs_sys_realloc). This mmaped
+     * for it (see ucs_callbackq_array_grow->ucs_sys_realloc). This mapped
      * address may conflict with the one used in this test, because
      * ucs::mmap_fixed_address() does mmap/munmap to obtain a pointer for
      * ucp_mem_map() with UCP_MEM_MAP_FIXED which creates a race with async
@@ -903,6 +955,77 @@ UCS_TEST_P(test_ucp_mmap, fixed) {
     }
 }
 
+UCS_TEST_P(test_ucp_mmap, gva_allocate, "GVA_ENABLE=y")
+{
+    for (auto mem_type : mem_buffer::supported_mem_types()) {
+        ucp_md_map_t md_map = sender().ucph()->gva_md_map[mem_type];
+        if (md_map == 0) {
+            continue;
+        }
+
+        ucp_mem_h memh;
+        ucp_mem_map_params_t params;
+        params.field_mask  = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
+                             UCP_MEM_MAP_PARAM_FIELD_LENGTH |
+                             UCP_MEM_MAP_PARAM_FIELD_FLAGS |
+                             UCP_MEM_MAP_PARAM_FIELD_MEMORY_TYPE;
+        params.address     = NULL;
+        params.memory_type = mem_type;
+        params.length      = 1 * UCS_MBYTE;
+        params.flags       = UCP_MEM_MAP_ALLOCATE;
+
+        ASSERT_UCS_OK(ucp_mem_map(sender().ucph(), &params, &memh));
+        EXPECT_TRUE(ucs_test_all_flags(memh->md_map, md_map));
+        ASSERT_UCS_OK(ucp_mem_unmap(sender().ucph(), memh));
+    }
+}
+
+UCS_TEST_P(test_ucp_mmap, gva, "GVA_ENABLE=y")
+{
+    std::list<void*> bufs;
+    ucp_mem_h first     = NULL;
+    ucp_md_map_t md_map = 0;
+
+    for (int i = 0; i < 1000 / ucs::test_time_multiplier(); ++i) {
+        size_t size  = (i + 1) * ((i % 2) ? 1000 : 1);
+        void *buffer = ucs_malloc(size, "gva");
+        bufs.push_back(buffer);
+
+        ucp_mem_h memh;
+        ucp_mem_map_params_t params;
+
+        params.field_mask = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
+                            UCP_MEM_MAP_PARAM_FIELD_LENGTH;
+        params.address    = buffer;
+        params.length     = size;
+
+        ASSERT_UCS_OK(ucp_mem_map(sender().ucph(), &params, &memh));
+
+        if (i == 0) {
+            first  = memh;
+            md_map = memh->md_map & sender().ucph()->gva_md_map[memh->mem_type];
+            if (md_map == 0) {
+                UCS_TEST_MESSAGE << "no GVA";
+                break;
+            }
+        } else {
+            ucp_md_index_t md_index;
+
+            ucs_for_each_bit(md_index, md_map) {
+                EXPECT_EQ(memh->uct[md_index], first->uct[md_index]);
+            }
+
+            ASSERT_UCS_OK(ucp_mem_unmap(sender().ucph(), memh));
+        }
+    }
+
+    ASSERT_UCS_OK(ucp_mem_unmap(sender().ucph(), first));
+    for (auto *buffer : bufs) {
+        ucs_free(buffer);
+    }
+}
+
+
 UCP_INSTANTIATE_TEST_CASE_GPU_AWARE(test_ucp_mmap)
 
 class test_ucp_mmap_atomic : public test_ucp_mmap {
@@ -1000,66 +1123,9 @@ public:
     }
 
 protected:
-    struct mem_chunk {
-        ucp_context_h           context;
-        ucp_mem_h               memh;
-        std::vector<ucp_rkey_h> rkeys;
-
-        mem_chunk(ucp_context_h);
-        ~mem_chunk();
-        ucp_rkey_h              unpack(ucp_ep_h, ucp_md_map_t md_map = 0);
-    };
-
     std::vector<std::unique_ptr<mem_chunk>> m_chunks;
 };
 
-test_ucp_rkey_compare::mem_chunk::mem_chunk(ucp_context_h ctx) : context(ctx)
-{
-    size_t size                 = 4096;
-    ucp_mem_map_params_t params = {
-        .field_mask = UCP_MEM_MAP_PARAM_FIELD_ADDRESS |
-                      UCP_MEM_MAP_PARAM_FIELD_LENGTH  |
-                      UCP_MEM_MAP_PARAM_FIELD_FLAGS,
-        .address    = NULL,
-        .length     = size,
-        .flags      = UCP_MEM_MAP_ALLOCATE,
-    };
-    ucs_status_t status;
-
-    status  = ucp_mem_map(context, &params, &memh);
-    ASSERT_UCS_OK(status);
-}
-
-test_ucp_rkey_compare::mem_chunk::~mem_chunk()
-{
-    for (auto &rkey : rkeys) {
-        ucp_rkey_destroy(rkey);
-    }
-
-    EXPECT_UCS_OK(ucp_mem_unmap(context, memh));
-}
-
-ucp_rkey_h
-test_ucp_rkey_compare::mem_chunk::unpack(ucp_ep_h ep, ucp_md_map_t md_map)
-{
-    ucp_rkey_h rkey;
-    void *rkey_buffer;
-    size_t rkey_size;
-
-    ASSERT_UCS_OK(ucp_rkey_pack(context, memh, &rkey_buffer, &rkey_size));
-    if (md_map == 0) {
-        ASSERT_UCS_OK(ucp_ep_rkey_unpack(ep, rkey_buffer, &rkey));
-    } else {
-        // Different MD map means different config index on proto v2
-        ASSERT_UCS_OK(ucp_ep_rkey_unpack_internal(ep, rkey_buffer, rkey_size,
-                                                  md_map, 0, &rkey));
-    }
-
-    ucp_rkey_buffer_release(rkey_buffer);
-    rkeys.push_back(rkey);
-    return rkey;
-}
-
 UCS_TEST_P(test_ucp_rkey_compare, rkey_compare_errors)
 {
     ucp_rkey_compare_params_t params = {};
@@ -1153,12 +1219,22 @@ public:
 
 UCS_TEST_P(test_ucp_mmap_export, reg_export_and_reimport)
 {
-    test_rereg(0, UCP_MEMH_PACK_FLAG_EXPORT, true);
+    test_rereg(0, true);
 }
 
 UCS_TEST_P(test_ucp_mmap_export, alloc_reg_export_and_reimport)
 {
-    test_rereg(UCP_MEM_MAP_ALLOCATE, UCP_MEMH_PACK_FLAG_EXPORT, true);
+    test_rereg(UCP_MEM_MAP_ALLOCATE, true);
+}
+
+UCS_TEST_P(test_ucp_mmap_export, export_import) {
+    mem_chunk mem(sender().ucph());
+    EXPECT_FALSE(mem.memh->flags & UCP_MEMH_FLAG_IMPORTED);
+
+    ucp_mem_h imported_memh = import_memh(mem.memh);
+    EXPECT_TRUE(imported_memh->flags & UCP_MEMH_FLAG_IMPORTED);
+
+    ASSERT_UCS_OK(ucp_mem_unmap(receiver().ucph(), imported_memh));
 }
 
 UCP_INSTANTIATE_TEST_CASE_GPU_AWARE(test_ucp_mmap_export)
