@@ -84,22 +84,6 @@ protected:
     bool ep_iface_has_caps(const entity& e, const std::string& tl,
                            uint64_t caps);
 
-    size_t count_resources(const ucp_test_base::entity &e,
-                           const std::string &tl_name) const
-    {
-        return std::count_if(e.ucph()->tl_rscs,
-                             e.ucph()->tl_rscs + e.ucph()->num_tls,
-                             [&](const ucp_tl_resource_desc_t &rsc) {
-                                 return tl_name == rsc.tl_rsc.tl_name;
-                             });
-    }
-
-    bool has_resource(const ucp_test_base::entity &e,
-                      const std::string &tl_name) const
-    {
-        return count_resources(e, tl_name) != 0;
-    }
-
 protected:
     vec_type                               m_send_data;
     vec_type                               m_recv_data;
@@ -311,7 +295,7 @@ void test_ucp_wireup::send_nb(ucp_ep_h ep, size_t length, int repeat,
         ucp_request_param_t param = {};
         void *req = ucp_ep_flush_nbx(ep, &param);
         ASSERT_UCS_PTR_OK(req);
-        reqs.push_back(req);
+        request_wait(req);
 
         req = ucp_put_nb(ep, &m_send_data[0], sizeof(m_send_data[0]),
                          (uintptr_t)&m_recv_data[length], rkey,
@@ -483,7 +467,7 @@ UCS_TEST_P(test_ucp_wireup_1sided, address) {
     EXPECT_LE(size, 2048ul); /* Expect a reasonable address size */
     EXPECT_EQ(addr_v, sender().ucph()->config.ext.worker_addr_version);
 
-    UCS_BITMAP_FOR_EACH_BIT(sender().worker()->context->tl_bitmap, tl) {
+    UCS_STATIC_BITMAP_FOR_EACH_BIT(tl, &sender().worker()->context->tl_bitmap) {
         const ucp_tl_resource_desc_t &rsc =
                 sender().worker()->context->tl_rscs[tl];
         packed_dev_priorities.insert(
@@ -1043,7 +1027,7 @@ public:
     bool check_scalable_tls(const ucp_worker_h worker, size_t est_num_eps) {
         ucp_rsc_index_t rsc_index;
 
-        UCS_BITMAP_FOR_EACH_BIT(worker->context->tl_bitmap, rsc_index) {
+        UCS_STATIC_BITMAP_FOR_EACH_BIT(rsc_index, &worker->context->tl_bitmap) {
             ucp_md_index_t md_index         = worker->context->tl_rscs[rsc_index].md_index;
             const uct_md_attr_v2_t *md_attr = &worker->context->tl_mds[md_index].attr;
 
@@ -1054,13 +1038,14 @@ public:
                 continue;
             }
 
-            if (ucp_worker_iface_get_attr(worker, rsc_index)->max_num_eps >= est_num_eps) {
-                EXPECT_TRUE(
-                        UCS_BITMAP_GET(worker->scalable_tl_bitmap, rsc_index));
+            if (ucp_worker_iface_get_attr(worker, rsc_index)->max_num_eps >=
+                est_num_eps) {
+                EXPECT_TRUE(UCS_STATIC_BITMAP_GET(worker->scalable_tl_bitmap,
+                                                  rsc_index));
                 return true;
             } else {
-                EXPECT_TRUE(UCS_BITMAP_GET(worker->scalable_tl_bitmap,
-                                           rsc_index) == 0);
+                EXPECT_FALSE(UCS_STATIC_BITMAP_GET(worker->scalable_tl_bitmap,
+                                                   rsc_index));
             }
         }
 
@@ -1396,8 +1381,8 @@ protected:
                 device_atomics_cnt++;
             }
         }
-        bool device_atomics_supported = !UCS_BITMAP_IS_ZERO_INPLACE(
-                &sender().worker()->atomic_tls);
+        bool device_atomics_supported = !UCS_STATIC_BITMAP_IS_ZERO(
+                sender().worker()->atomic_tls);
 
         test_ucp_wireup::cleanup();
 
@@ -1776,6 +1761,35 @@ public:
         double max_error = original / pow(2, _UCS_FP8_MANTISSA_BITS);
         EXPECT_NEAR(original, unpacked, max_error);
     }
+
+    const uct_iface_attr_t *get_iface_attr(const ucp_address_entry_t *ae)
+    {
+        ucp_worker_h worker   = sender().worker();
+        ucp_context_h context = worker->context;
+        ucp_rsc_index_t rsc_index;
+        uct_iface_is_reachable_params_t params;
+
+        params.field_mask  = UCT_IFACE_IS_REACHABLE_FIELD_DEVICE_ADDR |
+                             UCT_IFACE_IS_REACHABLE_FIELD_IFACE_ADDR |
+                             UCT_IFACE_IS_REACHABLE_FIELD_SCOPE;
+        params.device_addr = ae->dev_addr;
+        params.iface_addr  = ae->iface_addr;
+        params.scope       = UCT_IFACE_REACHABILITY_SCOPE_DEVICE;
+
+        UCS_STATIC_BITMAP_FOR_EACH_BIT(rsc_index, &context->tl_bitmap) {
+            auto wiface = ucp_worker_iface(worker, rsc_index);
+
+            /* Compare resources by device and transport */
+            if ((context->tl_rscs[rsc_index].tl_name_csum ==
+                 ae->tl_name_csum) &&
+                uct_iface_is_reachable_v2(wiface->iface, &params)) {
+                EXPECT_EQ(ae->md_index, context->tl_rscs[rsc_index].md_index);
+                return &wiface->attr;
+            }
+        }
+
+        return nullptr;
+    }
 };
 
 // On some systems TCP has very low BW and high latency, which would be
@@ -1806,8 +1820,8 @@ UCS_TEST_SKIP_COND_P(test_ucp_address_v2, pack_iface_attrs,
 
     const ucp_address_entry_t *ae;
     ucp_unpacked_address_for_each(ae, &unpacked_address) {
-        ucp_rsc_index_t rsc_idx = ae->iface_attr.dst_rsc_index;
-        uct_iface_attr_t *attr  = &ucp_worker_iface(worker, rsc_idx)->attr;
+        auto attr = get_iface_attr(ae);
+        ASSERT_NE(nullptr, attr);
 
         // Segment size is packed as a multiplicator of
         // UCP_ADDRESS_IFACE_SEG_SIZE_FACTOR, thus the unpacked value may be
