@@ -10,6 +10,11 @@ extern "C" {
 }
 #include <list>
 
+#if HAVE_MLX5_DV
+#include <uct/ib/base/ib_md.h>
+#include <uct/ib/mlx5/ib_mlx5.h>
+#endif
+
 class uct_flush_test : public uct_test {
 public:
     static const uint64_t SEED1 = 0x1111111111111111lu;
@@ -148,11 +153,18 @@ public:
     }
 
     void check_skip_test_flush_remote() {
-        if ((GetParam()->tl_name != "dc_mlx5") &&
-            (GetParam()->tl_name != "rc_verbs") &&
-            (GetParam()->tl_name != "rc_mlx5")) {
-            UCS_TEST_SKIP_R("not supported");
+#ifdef HAVE_MLX4_DV
+        auto md = sender().md();
+        if (std::string(md->component->name) == "ib") {
+            auto ib_md = ucs_derived_of(md, uct_ib_md_t);
+            if (ib_md->dev.flags & UCT_IB_DEVICE_FLAG_MLX5_PRM) {
+                auto mlx5_md = ucs_derived_of(ib_md, uct_ib_mlx5_md_t);
+                if (!(mlx5_md->flags & UCT_IB_MLX5_MD_FLAG_KSM)) {
+                    UCS_TEST_SKIP_R("not supported");
+                }
+            }
         }
+#endif
 
         m_flush_flags = UCT_FLUSH_FLAG_REMOTE;
     }
@@ -232,7 +244,8 @@ public:
     }
 
     void test_flush_put_zcopy(flush_func_t flush) {
-        const size_t length  = 10 * UCS_MBYTE;
+        const size_t length = ucs_min(10 * UCS_MBYTE,
+                                      sender().iface_attr().cap.put.max_zcopy);
         mapped_buffer sendbuf(length, SEED1, sender());
         mapped_buffer recvbuf(length, SEED2, receiver());
 
@@ -492,8 +505,8 @@ UCS_TEST_SKIP_COND_P(uct_flush_test, put_bcopy_flush_ep_remote,
 }
 
 UCS_TEST_SKIP_COND_P(uct_flush_test, put_zcopy_flush_ep_remote,
-                     !check_caps(UCT_IFACE_FLAG_PUT_BCOPY) ||
-                     !check_caps(UCT_IFACE_FLAG_GET_BCOPY)) {
+                     !check_caps(UCT_IFACE_FLAG_PUT_ZCOPY) ||
+                     !check_caps(UCT_IFACE_FLAG_GET_ZCOPY)) {
     check_skip_test_flush_remote();
     test_flush_put_zcopy(&uct_flush_test::flush_ep_nb);
 }
